import pytest

import numpy as np

import cupy
from cupy.cuda import driver
from cupy.cuda import runtime
from cupy import testing
import cupyx.scipy.signal  # NOQA

try:
    import scipy.signal  # NOQA
except ImportError:
    scipy = None


@pytest.mark.xfail(
    runtime.is_hip and driver.get_build_version() < 5_00_00000,
    reason='name_expressions with ROCm 4.3 may not work')
@testing.with_requires('scipy')
class TestLombscargle:
    @pytest.mark.parametrize('dtype', ['float32', 'float64'])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_frequency(self, dtype, xp, scp):
        """Test if frequency locations of peak corresponds to frequency of
        generated input signal.
        """
        dtype = xp.dtype(dtype)

        # Input parameters
        ampl = 2.
        w = 1.
        phi = 0.5 * xp.pi
        nin = 100
        nout = 1000
        p = 0.7  # Fraction of points to select

        # Randomly select a fraction of an array with timesteps
        r = testing.shaped_random((nin,), xp, scale=1.0,
                                  dtype=dtype, seed=2353425)
        t = xp.linspace(0.01 * xp.pi, 10. * xp.pi, nin, dtype=dtype)[r >= p]

        # Plot a sine wave for the selected times
        x = ampl * np.sin(w*t + phi)

        # Define the array of frequencies for which to compute the periodogram
        f = np.linspace(0.01, 10., nout, dtype=dtype)

        # Calculate Lomb-Scargle periodogram
        P = scp.signal.lombscargle(t, x, f)
        return P

    @pytest.mark.parametrize('dtype', ['float32', 'float64'])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_amplitude(self, dtype, xp, scp):
        # Test if height of peak in normalized Lomb-Scargle periodogram
        # corresponds to amplitude of the generated input signal.
        dtype = xp.dtype(dtype)

        # Input parameters
        ampl = 2.
        w = 1.
        phi = 0.5 * xp.pi
        nin = 100
        nout = 1000
        p = 0.7  # Fraction of points to select

        # Randomly select a fraction of an array with timesteps
        r = testing.shaped_random((nin,), xp, dtype=dtype, scale=1.0,
                                  seed=2353425)
        t = xp.linspace(0.01 * xp.pi, 10. * xp.pi, nin, dtype=dtype)[r >= p]

        # Plot a sine wave for the selected times
        x = ampl * xp.sin(w * t + phi)

        # Define the array of frequencies for which to compute the periodogram
        f = xp.linspace(0.01, 10., nout, dtype=dtype)

        # Calculate Lomb-Scargle periodogram
        pgram = scp.signal.lombscargle(t, x, f)

        # Normalize
        pgram = xp.sqrt(4 * pgram / t.shape[0])
        return pgram

    @pytest.mark.parametrize('dtype', ['float32', 'float64'])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_precenter(self, dtype, xp, scp):
        # Test if precenter gives the same result as manually precentering.
        dtype = xp.dtype(dtype)

        # Input parameters
        ampl = 2.
        w = 1.
        phi = 0.5 * xp.pi
        nin = 100
        nout = 1000
        p = 0.7  # Fraction of points to select
        offset = 0.15  # Offset to be subtracted in pre-centering

        # Randomly select a fraction of an array with timesteps
        r = testing.shaped_random((nin,), xp, dtype=dtype, scale=1.0,
                                  seed=2353425)
        t = xp.linspace(0.01 * xp.pi, 10. * xp.pi, nin, dtype=dtype)[r >= p]

        # Plot a sine wave for the selected times
        x = ampl * xp.sin(w * t + phi) + offset

        # Define the array of frequencies for which to compute the periodogram
        f = xp.linspace(0.01, 10., nout, dtype=dtype)

        # Calculate Lomb-Scargle periodogram
        pgram = scp.signal.lombscargle(t, x, f, precenter=True)
        pgram2 = scp.signal.lombscargle(t, x - x.mean(), f, precenter=False)

        # check if centering worked
        return pgram, pgram2

    @pytest.mark.parametrize('dtype', ['float32', 'float64'])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_normalize(self, dtype, xp, scp):
        # Test normalize option of Lomb-Scarge.

        # Input parameters
        ampl = 2.
        w = 1.
        phi = 0.5 * xp.pi
        nin = 100
        nout = 1000
        p = 0.7  # Fraction of points to select

        # Randomly select a fraction of an array with timesteps
        r = testing.shaped_random((nin,), xp, dtype=dtype, scale=1.0,
                                  seed=2353425)
        t = xp.linspace(0.01 * xp.pi, 10. * xp.pi, nin)[r >= p]

        # Plot a sine wave for the selected times
        x = ampl * xp.sin(w * t + phi)

        # Define the array of frequencies for which to compute the periodogram
        f = xp.linspace(0.01, 10., nout, dtype=dtype)

        # Calculate Lomb-Scargle periodogram
        pgram = scp.signal.lombscargle(t, x, f)
        pgram2 = scp.signal.lombscargle(t, x, f, normalize=True)

        # check if normalization works as expected
        return pgram, pgram2


@pytest.mark.xfail(
    runtime.is_hip and driver.get_build_version() < 5_00_00000,
    reason='name_expressions with ROCm 4.3 may not work')
@testing.with_requires('scipy')
class TestPeriodogram:
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_even(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        f, p = scp.signal.periodogram(x)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_odd(self, xp, scp):
        x = xp.zeros(15)
        x[0] = 1
        f, p = scp.signal.periodogram(x)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_twosided(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        f, p = scp.signal.periodogram(x, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_spectrum(self, xp, scp):
        x = np.zeros(16)
        x[0] = 1
        f, p = scp.signal.periodogram(x, scaling='spectrum')
        g, q = scp.signal.periodogram(x, scaling='density')
        return f, p, g, q

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_even(self, xp, scp):
        x = xp.zeros(16, dtype=int)
        x[0] = 1
        f, p = scp.signal.periodogram(x)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_odd(self, xp, scp):
        x = xp.zeros(15, dtype=int)
        x[0] = 1
        f, p = scp.signal.periodogram(x)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_twosided(self, xp, scp):
        x = np.zeros(16, dtype=int)
        x[0] = 1
        f, p = scp.signal.periodogram(x, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_complex(self, xp, scp):
        x = xp.zeros(16, xp.complex128)
        x[0] = 1.0 + 2.0j
        f, p = scp.signal.periodogram(x, return_onesided=False)
        return f, p

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_unk_scaling(self, mod):
        xp, scp = mod
        with pytest.raises(ValueError):
            x = xp.zeros(4, xp.complex128)
            scp.signal.periodogram(x, scaling='foo')

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nd_axis_m1(self, xp, scp):
        x = xp.zeros(20, dtype=np.float64)
        x = x.reshape((2, 1, 10))
        x[:, :, 0] = 1.0
        f, p = scp.signal.periodogram(x)
        f0, p0 = scp.signal.periodogram(x[0, 0, :])
        return f, p, f0, p0

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nd_axis_0(self, xp, scp):
        x = xp.zeros(20, dtype=np.float64)
        x = x.reshape((10, 2, 1))
        x[0, :, :] = 1.0
        f, p = scp.signal.periodogram(x, axis=0)
        f0, p0 = scp.signal.periodogram(x[:, 0, 0])
        return f, p, f0, p0

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_window_external(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        f, p = scp.signal.periodogram(x, 10, 'hann')
        win = scp.signal.get_window('hann', 16)
        fe, pe = scp.signal.periodogram(x, 10, win)

        win_err = scp.signal.get_window('hann', 32)
        with pytest.raises(ValueError):
            scp.signal.periodogram(x, 10, win_err)

        return f, p, fe, pe

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_padded_fft(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        f, p = scp.signal.periodogram(x)
        fp, pp = scp.signal.periodogram(x, nfft=32)
        return f, p, fp, pp

    @pytest.mark.parametrize('shape', [(0,), (3, 0), (0, 5, 2)])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_empty_input(self, shape, xp, scp):
        f, p = scp.signal.periodogram(xp.empty(shape))
        return f, p

    @pytest.mark.parametrize('shape', [(3, 0), (0, 5, 2)])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_empty_input_other_axis(self, shape, xp, scp):
        f, p = scp.signal.periodogram(xp.empty(shape), axis=1)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_short_nfft(self, xp, scp):
        x = xp.zeros(18)
        x[0] = 1
        f, p = scp.signal.periodogram(x, nfft=16)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nfft_is_xshape(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        f, p = scp.signal.periodogram(x, nfft=16)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_even_32(self, xp, scp):
        x = xp.zeros(16, 'f')
        x[0] = 1
        f, p = scp.signal.periodogram(x)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_odd_32(self, xp, scp):
        x = xp.zeros(15, 'f')
        x[0] = 1
        f, p = scp.signal.periodogram(x)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_twosided_32(self, xp, scp):
        x = xp.zeros(16, 'f')
        x[0] = 1
        f, p = scp.signal.periodogram(x, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_complex_32(self, xp, scp):
        x = xp.zeros(16, 'F')
        x[0] = 1.0 + 2.0j
        f, p = scp.signal.periodogram(x, return_onesided=False)
        return f, p

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_shorter_window_error(self, mod):
        xp, scp = mod
        x = xp.zeros(16)
        x[0] = 1
        win = scp.signal.get_window('hann', 10)
        with pytest.raises(ValueError):
            scp.signal.periodogram(x, window=win)


@testing.with_requires('scipy')
class TestWelch:
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_even(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=8)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_odd(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=9)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_twosided(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=8, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_spectrum(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=8, scaling='spectrum')
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_onesided_even(self, xp, scp):
        x = xp.zeros(16, dtype=int)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=8)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_onesided_odd(self, xp, scp):
        x = xp.zeros(16, dtype=int)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=9)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_twosided(self, xp, scp):
        x = xp.zeros(16, dtype=int)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=8, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_complex(self, xp, scp):
        x = xp.zeros(16, xp.complex128)
        x[0] = 1.0 + 2.0j
        x[8] = 1.0 + 2.0j
        f, p = scp.signal.welch(x, nperseg=8, return_onesided=False)
        return f, p

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_unk_scaling(self, mod):
        xp, scp = mod
        with pytest.raises(ValueError):
            x = xp.zeros(4, xp.complex128)
            scp.signal.welch(x, scaling='foo', nperseg=4)

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_detrend_linear(self, xp, scp):
        x = xp.arange(10, dtype=xp.float64) + 0.04
        f, p = scp.signal.welch(x, nperseg=10, detrend='linear')
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_no_detrending(self, xp, scp):
        x = xp.arange(10, dtype=xp.float64) + 0.04
        f1, p1 = scp.signal.welch(x, nperseg=10, detrend=False)
        f2, p2 = scp.signal.welch(x, nperseg=10, detrend=lambda x: x)
        return f1, p1, f2, p2

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_detrend_external(self, xp, scp):
        x = xp.arange(10, dtype=xp.float64) + 0.04
        f, p = scp.signal.welch(
            x, nperseg=10,
            detrend=lambda seg: scp.signal.detrend(seg, type='l'))
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_detrend_external_nd_m1(self, xp, scp):
        x = xp.arange(40, dtype=xp.float64) + 0.04
        x = x.reshape((2, 2, 10))
        f, p = scp.signal.welch(
            x, nperseg=10,
            detrend=lambda seg: scp.signal.detrend(seg, type='l'))
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_detrend_external_nd_0(self, xp, scp):
        x = xp.arange(20, dtype=xp.float64) + 0.04
        x = x.reshape((2, 1, 10))
        x = xp.moveaxis(x, 2, 0)
        f, p = scp.signal.welch(
            x, nperseg=10, axis=0,
            detrend=lambda seg: scp.signal.detrend(seg, axis=0, type='l'))
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nd_axis_m1(self, xp, scp):
        x = xp.arange(20, dtype=xp.float64) + 0.04
        x = x.reshape((2, 1, 10))
        f, p = scp.signal.welch(x, nperseg=10)
        f0, p0 = scp.signal.welch(x[0, 0, :], nperseg=10)
        return f, p, f0, p0

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nd_axis_0(self, xp, scp):
        x = xp.arange(20, dtype=xp.float64) + 0.04
        x = x.reshape((10, 2, 1))
        f, p = scp.signal.welch(x, nperseg=10, axis=0)
        f0, p0 = scp.signal.welch(x[:, 0, 0], nperseg=10)
        return f, p, f0, p0

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_window_external(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, 10, 'hann', nperseg=8)
        win = scp.signal.get_window('hann', 8)
        fe, pe = scp.signal.welch(x, 10, win, nperseg=None)

        with pytest.raises(ValueError):
            scp.signal.welch(x, 10, win, nperseg=4)

        with pytest.raises(ValueError):
            win_err = scp.signal.get_window('hann', 32)
            scp.signal.welch(x, 10, win_err, nperseg=None)

        return f, p, fe, pe

    @pytest.mark.parametrize('shape', [(0,), (3, 0), (0, 5, 2)])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_empty_input(self, shape, xp, scp):
        f, p = scp.signal.welch(xp.empty(shape))
        return f, p

    @pytest.mark.parametrize('shape', [(3, 0), (0, 5, 2)])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_empty_input_other_axis(self, shape, xp, scp):
        f, p = scp.signal.welch(xp.empty(shape), axis=1)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_short_data(self, xp, scp):
        x = np.zeros(8)
        x[0] = 1

        # default nperseg
        f, p = scp.signal.welch(x, window='hann')
        # user-specified nperseg
        f1, p1 = scp.signal.welch(x, window='hann', nperseg=256)
        # valid nperseg, doesn't give warning
        f2, p2 = scp.signal.welch(x, nperseg=8)
        return f, p, f1, p1, f2, p2

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_window_long_or_nd(self, mod):
        xp, scp = mod

        with pytest.raises(ValueError):
            scp.signal.welch(xp.zeros(4), 1, xp.array([1, 1, 1, 1, 1]))

        with pytest.raises(ValueError):
            scp.signal.welch(xp.zeros(4), 1, xp.arange(6).reshape((2, 3)))

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nondefault_noverlap(self, xp, scp):
        x = xp.zeros(64)
        x[::8] = 1
        f, p = scp.signal.welch(x, nperseg=16, noverlap=4)
        return f, p

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_bad_noverlap(self, mod):
        xp, scp = mod
        with pytest.raises(ValueError):
            scp.signal.welch(xp.zeros(4), 1, 'hann', 2, 7)

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_nfft_too_short(self, mod):
        xp, scp = mod
        with pytest.raises(ValueError):
            scp.signal.welch(xp.ones(12), nfft=3, nperseg=4)

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_even_32(self, xp, scp):
        x = xp.zeros(16, 'f')
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=8)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_odd_32(self, xp, scp):
        x = xp.zeros(16, 'f')
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=9)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_twosided_32(self, xp, scp):
        x = xp.zeros(16, 'f')
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=8, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_complex_32(self, xp, scp):
        x = xp.zeros(16, 'F')
        x[0] = 1.0 + 2.0j
        x[8] = 1.0 + 2.0j
        f, p = scp.signal.welch(x, nperseg=8, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_padded_freqs(self, xp, scp):
        x = xp.zeros(12)

        nfft = 24
        fodd1, _ = scp.signal.welch(x, nperseg=5, nfft=nfft)
        feven1, _ = scp.signal.welch(x, nperseg=6, nfft=nfft)

        nfft = 25
        fodd2, _ = scp.signal.welch(x, nperseg=5, nfft=nfft)
        feven2, _ = scp.signal.welch(x, nperseg=6, nfft=nfft)

        return fodd1, feven1, fodd2, feven2

    @pytest.mark.parametrize(
        'window', ['hann', 'bartlett', ('tukey', 0.1), 'flattop'])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_window_correction(self, window, xp, scp):
        A = 20
        fs = 1e4
        nperseg = int(fs // 10)
        fsig = 300

        tt = xp.arange(fs) / fs
        x = A * xp.sin(2 * np.pi * fsig * tt)

        _, p_spec = scp.signal.welch(
            x, fs=fs, nperseg=nperseg, window=window, scaling='spectrum')
        freq, p_dens = scp.signal.welch(
            x, fs=fs, nperseg=nperseg, window=window, scaling='density')

        return p_spec, freq, p_dens

    @pytest.mark.parametrize('axis', [0, 1, 2])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_axis_rolling(self, axis, xp, scp):
        x_flat = testing.shaped_random((1024,), xp)
        _, p_flat = scp.signal.welch(x_flat)

        newshape = [1,] * 3
        newshape[axis] = -1
        x = x_flat.reshape(newshape)

        _, p_plus = scp.signal.welch(x, axis=axis)  # Positive axis index
        # Negative axis index
        _, p_minus = scp.signal.welch(x, axis=axis - x.ndim)
        return p_flat, p_plus, p_minus

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_average(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.welch(x, nperseg=8, average='median')

        with pytest.raises(ValueError):
            scp.signal.welch(x, nperseg=8, average='unrecognised-average')
        return f, p


@testing.with_requires('scipy')
class TestCSD:
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_pad_shorter_x(self, xp, scp):
        x = xp.zeros(8)
        y = xp.zeros(12)
        f1, c1 = scp.signal.csd(x, y, nperseg=12)
        return f1, c1

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_pad_shorter_y(self, xp, scp):
        x = xp.zeros(12)
        y = xp.zeros(8)
        f1, c1 = scp.signal.csd(x, y, nperseg=12)
        return f1, c1

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_even(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=8)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_odd(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=9)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_twosided(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=8, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_spectrum(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=8, scaling='spectrum')
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_onesided_even(self, xp, scp):
        x = xp.zeros(16, dtype=int)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=8)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_onesided_odd(self, xp, scp):
        x = xp.zeros(16, dtype=int)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=9)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_integer_twosided(self, xp, scp):
        x = xp.zeros(16, dtype=int)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=8, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_complex(self, xp, scp):
        x = xp.zeros(16, xp.complex128)
        x[0] = 1.0 + 2.0j
        x[8] = 1.0 + 2.0j
        f, p = scp.signal.csd(x, x, nperseg=8, return_onesided=False)
        return f, p

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_unk_scaling(self, mod):
        xp, scp = mod
        with pytest.raises(ValueError):
            scp.signal.csd(
                xp.zeros(4, np.complex128), xp.ones(4, np.complex128),
                scaling='foo', nperseg=4)

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_detrend_linear(self, xp, scp):
        x = xp.arange(10, dtype=xp.float64) + 0.04
        f, p = scp.signal.csd(x, x, nperseg=10, detrend='linear')
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_no_detrending(self, xp, scp):
        x = xp.arange(10, dtype=xp.float64) + 0.04
        f1, p1 = scp.signal.csd(x, x, nperseg=10, detrend=False)
        f2, p2 = scp.signal.csd(x, x, nperseg=10, detrend=lambda x: x)
        return f1, p1, f2, p2

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_detrend_external(self, xp, scp):
        x = xp.arange(10, dtype=xp.float64) + 0.04
        f, p = scp.signal.csd(
            x, x, nperseg=10,
            detrend=lambda seg: scp.signal.detrend(seg, type='l'))
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_detrend_external_nd_m1(self, xp, scp):
        x = xp.arange(40, dtype=xp.float64) + 0.04
        x = x.reshape((2, 2, 10))
        f, p = scp.signal.csd(
            x, x, nperseg=10,
            detrend=lambda seg: scp.signal.detrend(seg, type='l'))
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_detrend_external_nd_0(self, xp, scp):
        x = xp.arange(20, dtype=xp.float64) + 0.04
        x = x.reshape((2, 1, 10))
        x = xp.moveaxis(x, 2, 0)
        f, p = scp.signal.csd(
            x, x, nperseg=10, axis=0,
            detrend=lambda seg: scp.signal.detrend(seg, axis=0, type='l'))
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nd_axis_m1(self, xp, scp):
        x = xp.arange(20, dtype=xp.float64) + 0.04
        x = x.reshape((2, 1, 10))
        f, p = scp.signal.csd(x, x, nperseg=10)
        f0, p0 = scp.signal.csd(x[0, 0, :], x[0, 0, :], nperseg=10)
        return f, p, f0, p0

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nd_axis_0(self, xp, scp):
        x = xp.arange(20, dtype=xp.float64) + 0.04
        x = x.reshape((10, 2, 1))
        f, p = scp.signal.csd(x, x, nperseg=10, axis=0)
        f0, p0 = scp.signal.csd(x[:, 0, 0], x[:, 0, 0], nperseg=10)
        return f, p, f0, p0

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_window_external(self, xp, scp):
        x = xp.zeros(16)
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, 10, 'hann', 8)
        win = scp.signal.get_window('hann', 8)
        fe, pe = scp.signal.csd(x, x, 10, win, nperseg=None)
        win_err = scp.signal.get_window('hann', 32)

        with pytest.raises(ValueError):
            scp.signal.csd(x, x, 10, win_err, nperseg=None)
        return f, p, fe, pe

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_empty_input(self, xp, scp):
        result = []

        for shape in [(0,), (3, 0), (0, 5, 2)]:
            f, p = scp.signal.csd(xp.empty(shape), xp.empty(shape))
            result += [f, p]

        f, p = scp.signal.csd(xp.ones(10), xp.empty((5, 0)))
        result += [f, p]

        f, p = scp.signal.csd(xp.empty((5, 0)), xp.ones(10))
        result += [f, p]
        return result

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_empty_input_other_axis(self, xp, scp):
        result = []
        for shape in [(3, 0), (0, 5, 2)]:
            f, p = scp.signal.csd(xp.empty(shape), xp.empty(shape), axis=1)
            result += [f, p]

        f, p = scp.signal.csd(xp.empty((10, 10, 3)),
                              xp.zeros((10, 0, 1)), axis=1)
        result += [f, p]

        f, p = scp.signal.csd(xp.empty((10, 0, 1)),
                              xp.zeros((10, 10, 3)), axis=1)
        result += [f, p]
        return result

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_short_data(self, xp, scp):
        x = xp.zeros(8)
        x[0] = 1

        # default nperseg
        f, p = scp.signal.csd(x, x, window='hann')
        # user-specified nperseg
        f1, p1 = scp.signal.csd(x, x, window='hann', nperseg=256)
        # valid nperseg, doesn't give warning
        f2, p2 = scp.signal.csd(x, x, nperseg=8)
        return f, p, f1, p1, f2, p2

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_window_long_or_nd(self, mod):
        xp, scp = mod
        with pytest.raises(ValueError):
            scp.signal.csd(xp.zeros(4), xp.zeros(
                4), 1, xp.array([1, 1, 1, 1, 1]))

        with pytest.raises(ValueError):
            scp.signal.csd(xp.zeros(4), xp.ones(4), 1,
                           xp.arange(6).reshape((2, 3)))

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_nondefault_noverlap(self, xp, scp):
        x = xp.zeros(64)
        x[::8] = 1
        f, p = scp.signal.csd(x, x, nperseg=16, noverlap=4)
        return f, p

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_bad_noverlap(self, mod):
        xp, scp = mod
        with pytest.raises(ValueError):
            scp.signal.csd(xp.zeros(4), xp.ones(4), 1, 'hann', 2, 7)

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_nfft_too_short(self, mod):
        xp, scp = mod
        with pytest.raises(ValueError):
            scp.signal.csd(xp.ones(12), xp.zeros(12), nfft=3,
                           nperseg=4)

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_even_32(self, xp, scp):
        x = xp.zeros(16, 'f')
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=8)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_onesided_odd_32(self, xp, scp):
        x = xp.zeros(16, 'f')
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=9)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_real_twosided_32(self, xp, scp):
        x = xp.zeros(16, 'f')
        x[0] = 1
        x[8] = 1
        f, p = scp.signal.csd(x, x, nperseg=8, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_complex_32(self, xp, scp):
        x = xp.zeros(16, 'F')
        x[0] = 1.0 + 2.0j
        x[8] = 1.0 + 2.0j
        f, p = scp.signal.csd(x, x, nperseg=8, return_onesided=False)
        return f, p

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_padded_freqs(self, xp, scp):
        x = xp.zeros(12)
        y = xp.ones(12)

        nfft = 24
        fodd1, _ = scp.signal.csd(x, y, nperseg=5, nfft=nfft)
        feven1, _ = scp.signal.csd(x, y, nperseg=6, nfft=nfft)

        nfft = 25
        fodd2, _ = scp.signal.csd(x, y, nperseg=5, nfft=nfft)
        feven2, _ = scp.signal.csd(x, y, nperseg=6, nfft=nfft)
        return fodd1, feven1, fodd2, feven2

    @pytest.mark.skipif(
        cupy.cuda.runtime.runtimeGetVersion() < 12000,
        reason="It fails on CUDA 11.x")
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_copied_data(self, xp, scp):
        x = testing.shaped_random((64,), xp, xp.float64)
        y = x.copy()

        _, p_same1 = scp.signal.csd(x, x, nperseg=8, average='mean',
                                    return_onesided=False)
        _, p_copied1 = scp.signal.csd(x, y, nperseg=8, average='mean',
                                      return_onesided=False)

        _, p_same2 = scp.signal.csd(x, x, nperseg=8, average='median',
                                    return_onesided=False)
        _, p_copied2 = scp.signal.csd(x, y, nperseg=8, average='median',
                                      return_onesided=False)
        return p_same1, p_copied1, p_same2, p_copied2


@testing.with_requires('scipy')
class TestCheckNOLACOLA:

    @pytest.mark.parametrize('setting', [
        ('boxcar', 10, 0),
        ('boxcar', 10, 9),
        ('bartlett', 51, 26),
        ('hann', 256, 128),
        ('hann', 256, 192),
        ('blackman', 300, 200),
        (('tukey', 0.5), 256, 64),
        ('hann', 256, 255),
    ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_check_COLA(self, setting, xp, scp):
        return scp.signal.check_COLA(*setting)

    @pytest.mark.parametrize('setting', [
        ('boxcar', 10, 0),
        ('boxcar', 10, 9),
        ('boxcar', 10, 7),
        ('bartlett', 51, 26),
        ('bartlett', 51, 10),
        ('hann', 256, 128),
        ('hann', 256, 192),
        ('hann', 256, 37),
        ('blackman', 300, 200),
        ('blackman', 300, 123),
        (('tukey', 0.5), 256, 64),
        (('tukey', 0.5), 256, 38),
        ('hann', 256, 255),
        ('hann', 256, 39),
    ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_check_NOLA(self, setting, xp, scp):
        return scp.signal.check_NOLA(*setting)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_check_NOLA_fail(self, xp, scp):
        w_fail = xp.ones(16)
        w_fail[::2] = 0

        settings_fail = [
            (w_fail, len(w_fail), len(w_fail) // 2),
            ('hann', 64, 0),
        ]
        result = []
        for setting in settings_fail:
            result.append(scp.signal.check_NOLA(*setting))
        return result


@testing.with_requires('scipy>=1.9.0')
class TestSTFT:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_average_all_segments(self, xp, scp):
        x = testing.shaped_random((1024,), xp, xp.float64, scale=1, seed=1234)

        fs = 1.0
        window = 'hann'
        nperseg = 16
        noverlap = 8

        # Compare twosided, because onesided welch doubles non-DC terms to
        # account for power at negative frequencies. stft doesn't do this,
        # because it breaks invertibility.
        f, _, Z = scp.signal.stft(
            x, fs, window, nperseg, noverlap, padded=False,
            return_onesided=False, boundary=None)

        return f, Z

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_permute_axes(self, xp, scp):
        x = testing.shaped_random((1024,), xp, xp.float64, scale=1, seed=1234)

        fs = 1.0
        window = 'hann'
        nperseg = 16
        noverlap = 8

        f1, t1, Z1 = scp.signal.stft(x, fs, window, nperseg, noverlap)
        f2, t2, Z2 = scp.signal.stft(
            x.reshape((-1, 1, 1)), fs, window, nperseg, noverlap, axis=0)

        t3, x1 = scp.signal.istft(Z1, fs, window, nperseg, noverlap)
        t4, x2 = scp.signal.istft(
            Z2.T, fs, window, nperseg, noverlap, time_axis=0, freq_axis=-1)

        return f1, t1, Z1, f2, t2, Z2, t3, x1, t4, x2.reshape(-1)

    @pytest.mark.parametrize('scaling', ['spectrum', 'psd'])
    @pytest.mark.parametrize('settings', [
        ('boxcar', 100, 10, 0),           # Test no overlap
        ('boxcar', 100, 10, 9),           # Test high overlap
        ('bartlett', 101, 51, 26),        # Test odd nperseg
        ('hann', 1024, 256, 128),         # Test defaults
        (('tukey', 0.5), 1152, 256, 64),  # Test Tukey
        ('hann', 1024, 256, 255),         # Test overlapped hann
    ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_roundtrip_real(self, scaling, settings, xp, scp):
        window, N, nperseg, noverlap = settings
        t = xp.arange(N)
        x = testing.shaped_random(t.shape, xp, xp.float64, seed=1234)

        _, _, zz = scp.signal.stft(
            x, nperseg=nperseg, noverlap=noverlap,
            window=window, detrend=None, padded=False,
            scaling=scaling)

        tr, xr = scp.signal.istft(
            zz, nperseg=nperseg, noverlap=noverlap, window=window,
            scaling=scaling)

        return zz, tr, xr

    @pytest.mark.parametrize('settings', [
        ('hann', 1024, 256, 128)
    ])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5, atol=1e-5)
    def test_roundtrip_float32(self, settings, xp, scp):
        window, N, nperseg, noverlap = settings
        t = xp.arange(N)
        x = testing.shaped_random(t.shape, xp, xp.float32, seed=1234)

        _, _, zz = scp.signal.stft(
            x, nperseg=nperseg, noverlap=noverlap,
            window=window, detrend=None, padded=False)

        tr, xr = scp.signal.istft(zz, nperseg=nperseg, noverlap=noverlap,
                                  window=window)

        return zz, tr, xr

    @pytest.mark.parametrize('scaling', ['spectrum', 'psd'])
    @pytest.mark.parametrize('settings', [
        ('boxcar', 100, 10, 0),           # Test no overlap
        ('boxcar', 100, 10, 9),           # Test high overlap
        ('bartlett', 101, 51, 26),        # Test odd nperseg
        ('hann', 1024, 256, 128),         # Test defaults
        (('tukey', 0.5), 1152, 256, 64),  # Test Tukey
        ('hann', 1024, 256, 255),         # Test overlapped hann
    ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_roundtrip_complex(self, scaling, settings, xp, scp):
        window, N, nperseg, noverlap = settings
        t = np.arange(N)
        x_real = testing.shaped_random(t.shape, xp, xp.float64, seed=1234)
        x_imag = 1j * testing.shaped_random(t.shape, xp, xp.float64)
        x = x_real + x_imag

        _, _, zz = scp.signal.stft(
            x, nperseg=nperseg, noverlap=noverlap,
            window=window, detrend=None, padded=False,
            return_onesided=False, scaling=scaling)

        tr, xr = scp.signal.istft(
            zz, nperseg=nperseg, noverlap=noverlap,
            window=window, input_onesided=False, scaling=scaling)

        return zz, tr, xr

    @pytest.mark.parametrize('settings', [
        ('boxcar', 100, 10, 0),           # Test no overlap
        ('boxcar', 100, 10, 9),           # Test high overlap
    ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_roundtrip_boundary_extension(self, settings, xp, scp):
        # Test against boxcar, since window is all ones, and thus can be fully
        # recovered with no boundary extension
        window, N, nperseg, noverlap = settings
        t = xp.arange(N)
        x = testing.shaped_random(t.shape, xp, xp.float64, seed=1234)

        results = []
        _, _, zz = scp.signal.stft(
            x, nperseg=nperseg, noverlap=noverlap,
            window=window, detrend=None, padded=True, boundary=None)

        _, xr = scp.signal.istft(
            zz, noverlap=noverlap, window=window, boundary=False)

        results.append(zz)
        results.append(xr)

        for boundary in ['even', 'odd', 'constant', 'zeros']:
            _, _, zz_ext = scp.signal.stft(
                x, nperseg=nperseg, noverlap=noverlap,
                window=window, detrend=None, padded=True,
                boundary=boundary)

            _, xr_ext = scp.signal.istft(
                zz_ext, noverlap=noverlap, window=window, boundary=True)

            results.append(zz_ext)
            results.append(xr_ext)

        return results

    @pytest.mark.parametrize('settings', [
        ('boxcar', 101, 10, 0),
        ('hann', 1000, 256, 128),
    ])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-7)
    def test_roundtrip_padded_signal(self, settings, xp, scp):
        window, N, nperseg, noverlap = settings
        t = xp.arange(N)
        x = testing.shaped_random(t.shape, xp, xp.float64, seed=1234)

        _, _, zz = scp.signal.stft(
            x, nperseg=nperseg, noverlap=noverlap,
            window=window, detrend=None, padded=True)

        tr, xr = scp.signal.istft(zz, noverlap=noverlap, window=window)
        return zz, tr, xr

    @pytest.mark.parametrize('settings', [
        ('hann', 1024, 256, 128, 512),
        ('hann', 1024, 256, 128, 501),
        ('boxcar', 100, 10, 0, 33),
        (('tukey', 0.5), 1152, 256, 64, 1024),
    ])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_roundtrip_padded_FFT(self, settings, xp, scp):
        window, N, nperseg, noverlap, nfft = settings
        t = np.arange(N)
        x = testing.shaped_random(t.shape, xp, xp.float64, seed=1234)
        xc = x * xp.exp(1j * xp.pi / 4)

        # real signal
        _, _, z = scp.signal.stft(
            x, nperseg=nperseg, noverlap=noverlap, nfft=nfft,
            window=window, detrend=None, padded=True)

        # complex signal
        _, _, zc = scp.signal.stft(
            xc, nperseg=nperseg, noverlap=noverlap, nfft=nfft,
            window=window, detrend=None, padded=True, return_onesided=False)

        tr, xr = scp.signal.istft(
            z, nperseg=nperseg, noverlap=noverlap, nfft=nfft, window=window)

        tcr, xcr = scp.signal.istft(
            zc, nperseg=nperseg, noverlap=noverlap, nfft=nfft,
            window=window, input_onesided=False)
        return z, zc, xr, xcr, tr, tcr

    @pytest.mark.parametrize('a', [0, 1, 2])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_axis_rolling(self, a, xp, scp):
        x_flat = testing.shaped_random((1024,), xp, xp.float64, seed=1234)
        _, _, z_flat = scp.signal.stft(x_flat)

        newshape = [1,]*3
        newshape[a] = -1
        x = x_flat.reshape(newshape)

        # Positive axis index
        _, _, z_plus = scp.signal.stft(x, axis=a)
        # Negative axis index
        _, _, z_minus = scp.signal.stft(x, axis=a - x.ndim)

        # z_flat has shape [n_freq, n_time]

        # Test vs. transpose
        _, x_transpose_m = scp.signal.istft(
            z_flat.T, time_axis=-2, freq_axis=-1)
        _, x_transpose_p = scp.signal.istft(
            z_flat.T, time_axis=0, freq_axis=1)
        return z_plus, z_minus, x_transpose_m, x_transpose_p

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-7)
    def test_roundtrip_scaling(self, xp, scp):
        """Verify behavior of scaling parameter. """
        # Create 1024 sample cosine signal with amplitude 2:
        X = xp.zeros(513, dtype=complex)
        X[256] = 1024
        x = xp.fft.irfft(X)

        results = []

        # Calculate magnitude-scaled STFT:
        Zs = scp.signal.stft(x, boundary='even', scaling='spectrum')[2]
        results.append(Zs)

        # Test round trip:
        x1 = scp.signal.istft(Zs, boundary=True, scaling='spectrum')[1]
        results.append(x1)

        # Calculate two-sided psd-scaled STFT:
        #  - using 'even' padding since signal is axis symmetric - this ensures
        #    stationary behavior on the boundaries
        #  - using the two-sided transform allows determining the spectral
        #    power by `sum(abs(Zp[:, k])**2) / len(f)` for the k-th time slot.
        Zp = scp.signal.stft(
            x, return_onesided=False, boundary='even', scaling='psd')[2]
        results.append(Zp)

        # Test round trip:
        x1 = scp.signal.istft(
            Zp, input_onesided=False, boundary=True, scaling='psd')[1]
        results.append(x1)

        # The power of the one-sided psd-scaled STFT can be determined
        # analogously (note that the two sides are not of equal shape):
        Zp0 = scp.signal.stft(
            x, return_onesided=True, boundary='even', scaling='psd')[2]
        results.append(Zp0)

        # Test round trip:
        x1 = scp.signal.istft(
            Zp0, input_onesided=True, boundary=True, scaling='psd')[1]
        results.append(x1)
        return results


@testing.with_requires('scipy')
class TestVectorstrength:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_single_1dperiod(self, xp, scp):
        events = xp.array([.5])
        period = 5.

        strength, phase = scp.signal.vectorstrength(events, period)
        return strength, phase

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_single_2dperiod(self, xp, scp):
        events = xp.array([.5])
        period = [1, 2, 5.]

        strength, phase = scp.signal.vectorstrength(events, period)
        return strength, phase

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_equal_1dperiod(self, xp, scp):
        events = xp.array([.25, .25, .25, .25, .25, .25])
        period = 2
        strength, phase = scp.signal.vectorstrength(events, period)
        return strength, phase

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_equal_2dperiod(self, xp, scp):
        events = xp.array([.25, .25, .25, .25, .25, .25])
        period = [1, 2, ]

        strength, phase = scp.signal.vectorstrength(events, period)
        return strength, phase

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_spaced_1dperiod(self, xp, scp):
        events = xp.array([.1, 1.1, 2.1, 4.1, 10.1])
        period = 1

        strength, phase = scp.signal.vectorstrength(events, period)
        return strength, phase

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_spaced_2dperiod(self, xp, scp):
        events = xp.array([.1, 1.1, 2.1, 4.1, 10.1])
        period = [1, .5]

        strength, phase = scp.signal.vectorstrength(events, period)
        return strength, phase

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_partial_1dperiod(self, xp, scp):
        events = xp.array([.25, .5, .75])
        period = 1

        strength, phase = scp.signal.vectorstrength(events, period)
        return strength, phase

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_partial_2dperiod(self, xp, scp):
        events = xp.array([.25, .5, .75])
        period = [1., 1., 1., 1.]

        strength, phase = scp.signal.vectorstrength(events, period)
        return strength, phase

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-7, atol=1e-7)
    def test_opposite_1dperiod(self, xp, scp):
        events = xp.array([0, .25, .5, .75])
        period = 1.
        strength, _ = scp.signal.vectorstrength(events, period)
        return strength

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-7, atol=1e-7)
    def test_opposite_2dperiod(self, xp, scp):
        events = xp.array([0, .25, .5, .75])
        period = [1.] * 10
        strength, _ = scp.signal.vectorstrength(events, period)
        return strength

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_2d_events_ValueError(self, mod):
        xp, scp = mod
        events = xp.array([[1, 2]])
        period = 1.
        with pytest.raises(ValueError):
            scp.signal.vectorstrength(events, period)

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_2d_period_ValueError(self, mod):
        xp, scp = mod
        events = 1.
        period = xp.array([[1]])
        with pytest.raises(ValueError):
            scp.signal.vectorstrength(events, period)

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_zero_period_ValueError(self, mod):
        _, scp = mod
        events = 1.
        period = 0
        with pytest.raises(ValueError):
            scp.signal.vectorstrength(events, period)

    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_negative_period_ValueError(self, mod):
        _, scp = mod
        events = 1.
        period = -1
        with pytest.raises(ValueError):
            scp.signal.vectorstrength(events, period)


@testing.with_requires('scipy')
class TestCoherence:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_identical_input(self, xp, scp):
        x = testing.shaped_random((20,), xp, xp.float64, scale=1.0)
        y = xp.copy(x)  # So `y is x` -> False

        f1, C1 = scp.signal.coherence(x, y, nperseg=10)
        return f1, C1

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_phase_shifted_input(self, xp, scp):
        x = testing.shaped_random((20,), xp, xp.float64, scale=1.0)
        y = -x

        f1, C1 = scp.signal.coherence(x, y, nperseg=10)
        return f1, C1


@testing.with_requires('scipy')
class TestSpectrogram:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_average_all_segments(self, xp, scp):
        x = testing.shaped_random((1024,), xp, xp.float64, scale=1.0)

        fs = 1.0
        window = ('tukey', 0.25)
        nperseg = 16
        noverlap = 2

        f, _, P = scp.signal.spectrogram(x, fs, window, nperseg, noverlap)
        return f, P

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_window_external(self, xp, scp):
        x = testing.shaped_random((1024,), xp, xp.float64, scale=1.0)

        fs = 1.0
        win = scp.signal.get_window(('tukey', 0.25), 16)
        fe, _, Pe = scp.signal.spectrogram(
            x, fs, win, nperseg=None, noverlap=2)

        with pytest.raises(ValueError):
            scp.signal.spectrogram(x, fs, win, nperseg=8)

        win_err = scp.signal.get_window(('tukey', 0.25), 2048)
        with pytest.raises(ValueError):
            scp.signal.spectrogram(x, fs, win_err, nperseg=None)
        return fe, Pe

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_short_data(self, xp, scp):
        x = testing.shaped_random((1024,), xp, xp.float64, scale=1.0)
        fs = 1.0

        # for string-like window, input signal length < nperseg value gives
        # UserWarning, sets nperseg to x.shape[-1]
        # default nperseg
        f, _, p = scp.signal.spectrogram(x, fs, window=('tukey', 0.25))

        # user-specified nperseg
        f1, _, p1 = scp.signal.spectrogram(
            x, fs, window=('tukey', 0.25), nperseg=1025)
        # to compare w/default
        f2, _, p2 = scp.signal.spectrogram(x, fs, nperseg=256)
        # compare w/user-spec'd
        f3, _, p3 = scp.signal.spectrogram(x, fs, nperseg=1024)
        return f, p, f1, p1, f2, p2, f3, p3



cdef list _elementwise_accelerators

cdef list _reduction_accelerators

cdef list _routine_accelerators

cpdef enum:
    ACCELERATOR_CUB = 1
    ACCELERATOR_CUTENSOR = 2
    ACCELERATOR_CUTENSORNET = 3



import math
import os

import cupy
import numpy as np

from ._util import _get_inttype
from ._pba_2d import (_check_distances, _check_indices,
                      _distance_tranform_arg_check, _generate_indices_ops,
                      _generate_shape, _get_block_size, lcm)

pba3d_defines_template = """

#define MARKER     {marker}
#define MAX_INT    {max_int}
#define BLOCKSIZE  {block_size_3d}

"""

# For efficiency, the original PBA+ packs three 10-bit integers and two binary
# flags into a single 32-bit integer. The defines in
# `pba3d_defines_encode_32bit` handle this format.
pba3d_defines_encode_32bit = """
// Sites     : ENCODE(x, y, z, 0, 0)
// Not sites : ENCODE(0, 0, 0, 1, 0) or MARKER
#define ENCODED_INT_TYPE int
#define ZERO 0
#define ONE 1
#define ENCODE(x, y, z, a, b)  (((x) << 20) | ((y) << 10) | (z) | ((a) << 31) | ((b) << 30))
#define DECODE(value, x, y, z) \
    x = ((value) >> 20) & 0x3ff; \
    y = ((value) >> 10) & 0x3ff; \
    z = (value) & 0x3ff

#define NOTSITE(value)  (((value) >> 31) & 1)
#define HASNEXT(value)  (((value) >> 30) & 1)

#define GET_X(value)    (((value) >> 20) & 0x3ff)
#define GET_Y(value)    (((value) >> 10) & 0x3ff)
#define GET_Z(value)    ((NOTSITE((value))) ? MAX_INT : ((value) & 0x3ff))

"""  # noqa


# 64bit version of ENCODE/DECODE to allow a 20-bit integer per coordinate axis.
pba3d_defines_encode_64bit = """
// Sites     : ENCODE(x, y, z, 0, 0)
// Not sites : ENCODE(0, 0, 0, 1, 0) or MARKER
#define ENCODED_INT_TYPE long long
#define ZERO 0L
#define ONE 1L
#define ENCODE(x, y, z, a, b)  (((x) << 40) | ((y) << 20) | (z) | ((a) << 61) | ((b) << 60))
#define DECODE(value, x, y, z) \
    x = ((value) >> 40) & 0xfffff; \
    y = ((value) >> 20) & 0xfffff; \
    z = (value) & 0xfffff

#define NOTSITE(value)  (((value) >> 61) & 1)
#define HASNEXT(value)  (((value) >> 60) & 1)

#define GET_X(value)    (((value) >> 40) & 0xfffff)
#define GET_Y(value)    (((value) >> 20) & 0xfffff)
#define GET_Z(value)    ((NOTSITE((value))) ? MAX_INT : ((value) & 0xfffff))

"""  # noqa


@cupy.memoize(True)
def get_pba3d_src(block_size_3d=32, marker=-2147483648, max_int=2147483647,
                  size_max=1024):
    pba3d_code = pba3d_defines_template.format(
        block_size_3d=block_size_3d, marker=marker, max_int=max_int
    )
    if size_max > 1024:
        pba3d_code += pba3d_defines_encode_64bit
    else:
        pba3d_code += pba3d_defines_encode_32bit
    kernel_directory = os.path.join(os.path.dirname(__file__), "cuda")
    with open(os.path.join(kernel_directory, "pba_kernels_3d.h"), "rt") as f:
        pba3d_kernels = "\n".join(f.readlines())
    pba3d_code += pba3d_kernels
    return pba3d_code


@cupy.memoize(for_each_device=True)
def _get_encode3d_kernel(size_max, marker=-2147483648):
    """Pack array coordinates into a single integer."""
    if size_max > 1024:
        int_type = "ptrdiff_t"  # int64_t
    else:
        int_type = "int"        # int32_t

    # value must match TOID macro in the C++ code!
    if size_max > 1024:
        value = """(((x) << 40) | ((y) << 20) | (z))"""
    else:
        value = """(((x) << 20) | ((y) << 10) | (z))"""

    code = f"""
    if (arr[i]) {{
        out[i] = {marker};
    }} else {{
        {int_type} shape_2 = arr.shape()[2];
        {int_type} shape_1 = arr.shape()[1];
        {int_type} _i = i;
        {int_type} x = _i % shape_2;
        _i /= shape_2;
        {int_type} y = _i % shape_1;
        _i /= shape_1;
        {int_type} z = _i;
        out[i] = {value};
    }}
    """
    return cupy.ElementwiseKernel(
        in_params="raw B arr",
        out_params="raw I out",
        operation=code,
        options=("--std=c++11",),
    )


def encode3d(arr, marker=-2147483648, bit_depth=32, size_max=1024):
    if arr.ndim != 3:
        raise ValueError("only 3d arr supported")
    if bit_depth not in [32, 64]:
        raise ValueError("only bit_depth of 32 or 64 is supported")
    if size_max > 1024:
        dtype = np.int64
    else:
        dtype = np.int32
    image = cupy.zeros(arr.shape, dtype=dtype, order="C")
    kern = _get_encode3d_kernel(size_max, marker=marker)
    kern(arr, image, size=image.size)
    return image


def _get_decode3d_code(size_max, int_type=""):
    # bit shifts here must match those used in the encode3d kernel
    if size_max > 1024:
        code = f"""
        {int_type} x = (encoded >> 40) & 0xfffff;
        {int_type} y = (encoded >> 20) & 0xfffff;
        {int_type} z = encoded & 0xfffff;
        """
    else:
        code = f"""
        {int_type} x = (encoded >> 20) & 0x3ff;
        {int_type} y = (encoded >> 10) & 0x3ff;
        {int_type} z = encoded & 0x3ff;
        """
    return code


@cupy.memoize(for_each_device=True)
def _get_decode3d_kernel(size_max):
    """Unpack 3 coordinates encoded as a single integer."""

    # int_type = "" here because x, y, z were already allocated externally
    code = _get_decode3d_code(size_max, int_type="")

    return cupy.ElementwiseKernel(
        in_params="E encoded",
        out_params="I x, I y, I z",
        operation=code,
        options=("--std=c++11",),
    )


def decode3d(encoded, size_max=1024):
    coord_dtype = cupy.int32 if size_max < 2**31 else cupy.int64
    x = cupy.empty_like(encoded, dtype=coord_dtype)
    y = cupy.empty_like(x)
    z = cupy.empty_like(x)
    kern = _get_decode3d_kernel(size_max)
    kern(encoded, x, y, z)
    return (x, y, z)


def _determine_padding(shape, block_size, m1, m2, m3, blockx, blocky):
    # TODO: can possibly revise to consider only particular factors for LCM on
    #       a given axis
    LCM = lcm(block_size, m1, m2, m3, blockx, blocky)
    orig_sz, orig_sy, orig_sx = shape
    round_up = False
    if orig_sx % LCM != 0:
        # round up size to a multiple of the band size
        round_up = True
        sx = LCM * math.ceil(orig_sx / LCM)
    else:
        sx = orig_sx
    if orig_sy % LCM != 0:
        # round up size to a multiple of the band size
        round_up = True
        sy = LCM * math.ceil(orig_sy / LCM)
    else:
        sy = orig_sy
    if orig_sz % LCM != 0:
        # round up size to a multiple of the band size
        round_up = True
        sz = LCM * math.ceil(orig_sz / LCM)
    else:
        sz = orig_sz

    aniso = not (sx == sy == sz)
    if aniso or round_up:
        smax = max(sz, sy, sx)
        padding_width = (
            (0, smax - orig_sz), (0, smax - orig_sy), (0, smax - orig_sx)
        )
    else:
        padding_width = None
    return padding_width


def _generate_distance_computation(int_type, dist_int_type):
    """
    Compute euclidean distance from current coordinate (ind_0, ind_1, ind_2) to
    the coordinates of the nearest point (z, y, x)."""
    return f"""
    {int_type} tmp = z - ind_0;
    {dist_int_type} sq_dist = tmp * tmp;
    tmp = y - ind_1;
    sq_dist += tmp * tmp;
    tmp = x - ind_2;
    sq_dist += tmp * tmp;
    dist[i] = sqrt(static_cast<F>(sq_dist));
    """


def _get_distance_kernel_code(int_type, dist_int_type, raw_out_var=True):
    code = _generate_shape(
        ndim=3, int_type=int_type, var_name="dist", raw_var=raw_out_var
    )
    code += _generate_indices_ops(ndim=3, int_type=int_type)
    code += _generate_distance_computation(int_type, dist_int_type)
    return code


@cupy.memoize(for_each_device=True)
def _get_distance_kernel(int_type, large_dist=False):
    """Returns kernel computing the Euclidean distance from coordinates."""
    dist_int_type = "ptrdiff_t" if large_dist else "int"
    operation = _get_distance_kernel_code(
        int_type, dist_int_type, raw_out_var=True
    )
    return cupy.ElementwiseKernel(
        in_params="I z, I y, I x",
        out_params="raw F dist",
        operation=operation,
        options=("--std=c++11",),
    )


def _generate_aniso_distance_computation():
    """
    Compute euclidean distance from current coordinate (ind_0, ind_1, ind_2) to
    the coordinates of the nearest point (z, y, x)."""
    return """
    F tmp = static_cast<F>(z - ind_0) * sampling[0];
    F sq_dist = tmp * tmp;
    tmp = static_cast<F>(y - ind_1) * sampling[1];
    sq_dist += tmp * tmp;
    tmp = static_cast<F>(x - ind_2) * sampling[2];
    sq_dist += tmp * tmp;
    dist[i] = sqrt(static_cast<F>(sq_dist));
    """


def _get_aniso_distance_kernel_code(int_type, raw_out_var=True):
    code = _generate_shape(
        ndim=3, int_type=int_type, var_name="dist", raw_var=raw_out_var
    )
    code += _generate_indices_ops(ndim=3, int_type=int_type)
    code += _generate_aniso_distance_computation()
    return code


@cupy.memoize(for_each_device=True)
def _get_aniso_distance_kernel(int_type):
    """Returns kernel computing the Euclidean distance from coordinates with
    axis spacing != 1."""
    operation = _get_aniso_distance_kernel_code(
        int_type, raw_out_var=True
    )
    return cupy.ElementwiseKernel(
        in_params="I z, I y, I x, raw F sampling",
        out_params="raw F dist",
        operation=operation,
        options=("--std=c++11",),
    )


@cupy.memoize(for_each_device=True)
def _get_decode_as_distance_kernel(size_max, large_dist=False, sampling=None):
    """Fused decode3d and distance computation.

    This kernel is for use when `return_distances=True`, but
    `return_indices=False`. It replaces the separate calls to
    `_get_decode3d_kernel` and `_get_distance_kernel`, avoiding the overhead of
    generating full arrays containing the coordinates since the coordinate
    arrays are not going to be returned.
    """
    if sampling is None:
        dist_int_type = "ptrdiff_t" if large_dist else "int"
    int_type = "int"

    # Step 1: decode the (z, y, x) coordinate
    code = _get_decode3d_code(size_max, int_type=int_type)

    # Step 2: compute the Euclidean distance based on this (z, y, x).
    code += _generate_shape(
        ndim=3, int_type=int_type, var_name="dist", raw_var=True
    )
    code += _generate_indices_ops(ndim=3, int_type=int_type)
    if sampling is None:
        code += _generate_distance_computation(int_type, dist_int_type)
        in_params = "E encoded"
    else:
        code += _generate_aniso_distance_computation()
        in_params = "E encoded, raw F sampling"
    return cupy.ElementwiseKernel(
        in_params=in_params,
        out_params="raw F dist",
        operation=code,
        options=("--std=c++11",),
    )


def _pba_3d(arr, sampling=None, return_distances=True, return_indices=False,
            block_params=None, check_warp_size=False, *,
            float64_distances=False, distances=None, indices=None):

    indices_inplace = isinstance(indices, cupy.ndarray)
    dt_inplace = isinstance(distances, cupy.ndarray)
    _distance_tranform_arg_check(
        dt_inplace, indices_inplace, return_distances, return_indices
    )

    if arr.ndim != 3:
        raise ValueError(f"expected a 3D array, got {arr.ndim}D")

    if block_params is None:
        m1 = 1
        m2 = 1
        m3 = 2
    else:
        m1, m2, m3 = block_params

    # reduce blockx for small inputs
    s_min = min(arr.shape)
    if s_min <= 4:
        blockx = 4
    elif s_min <= 8:
        blockx = 8
    elif s_min <= 16:
        blockx = 16
    else:
        blockx = 32
    blocky = 4

    block_size = _get_block_size(check_warp_size)

    orig_sz, orig_sy, orig_sx = arr.shape
    padding_width = _determine_padding(
        arr.shape, block_size, m1, m2, m3, blockx, blocky
    )
    if padding_width is not None:
        arr = cupy.pad(arr, padding_width, mode="constant", constant_values=1)
    size = arr.shape[0]

    # pba algorithm was implemented to use 32-bit integer to store compressed
    # coordinates. input_arr will be C-contiguous, int32
    size_max = max(arr.shape)
    input_arr = encode3d(arr, size_max=size_max)
    buffer_idx = 0
    output = cupy.zeros_like(input_arr)
    pba_images = [input_arr, output]

    block = (blockx, blocky, 1)
    grid = (size // block[0], size // block[1], 1)
    pba3d = cupy.RawModule(
        code=get_pba3d_src(block_size_3d=block_size, size_max=size_max)
    )

    kernelFloodZ = pba3d.get_function("kernelFloodZ")
    if sampling is None:
        kernelMaurerAxis = pba3d.get_function("kernelMaurerAxis")
        kernelColorAxis = pba3d.get_function("kernelColorAxis")
        sampling_args = ()
    else:
        kernelMaurerAxis = pba3d.get_function("kernelMaurerAxisWithSpacing")
        kernelColorAxis = pba3d.get_function("kernelColorAxisWithSpacing")
        sampling = tuple(map(float, sampling))
        sampling_args = (sampling[2], sampling[1], sampling[0])

    kernelFloodZ(
        grid,
        block,
        (pba_images[buffer_idx], pba_images[1 - buffer_idx], size)
    )
    buffer_idx = 1 - buffer_idx

    block = (blockx, blocky, 1)
    grid = (size // block[0], size // block[1], 1)
    kernelMaurerAxis(
        grid,
        block,
        (pba_images[buffer_idx], pba_images[1 - buffer_idx], size) + sampling_args,  # noqa
    )

    block = (block_size, m3, 1)
    grid = (size // block[0], size, 1)
    kernelColorAxis(
        grid,
        block,
        (pba_images[1 - buffer_idx], pba_images[buffer_idx], size) + sampling_args,  # noqa
    )

    if sampling is not None:
        # kernelColorAxis transposes the first two axis, so have to reorder
        # the sampling_args tuple correspondingly
        sampling_args = (sampling[1], sampling[2], sampling[0])

    block = (blockx, blocky, 1)
    grid = (size // block[0], size // block[1], 1)
    kernelMaurerAxis(
        grid,
        block,
        (pba_images[buffer_idx], pba_images[1 - buffer_idx], size) + sampling_args,  # noqa
    )

    block = (block_size, m3, 1)
    grid = (size // block[0], size, 1)
    kernelColorAxis(
        grid,
        block,
        (pba_images[1 - buffer_idx], pba_images[buffer_idx], size) + sampling_args,  # noqa
    )
    output = pba_images[buffer_idx]

    if return_distances:
        out_shape = (orig_sz, orig_sy, orig_sx)
        dtype_out = cupy.float64 if float64_distances else cupy.float32
        if dt_inplace:
            _check_distances(distances, out_shape, dtype_out)
        else:
            distances = cupy.zeros(out_shape, dtype=dtype_out)

        # make sure maximum possible distance doesn't overflow
        max_possible_dist = sum((s - 1)**2 for s in out_shape)
        large_dist = max_possible_dist >= 2**31

        if not return_indices:
            # Compute distances without forming explicit coordinate arrays.
            kern = _get_decode_as_distance_kernel(
                size_max=size_max,
                large_dist=large_dist,
                sampling=sampling
            )
            if sampling is None:
                kern(output[:orig_sz, :orig_sy, :orig_sx], distances)
            else:
                sampling = cupy.asarray(sampling, dtype=distances.dtype)
                kern(output[:orig_sz, :orig_sy, :orig_sx], sampling, distances)
            return (distances,)

    if return_indices:
        x, y, z = decode3d(output[:orig_sz, :orig_sy, :orig_sx],
                           size_max=size_max)
    vals = ()
    if return_distances:
        if sampling is None:
            kern = _get_distance_kernel(
                int_type=_get_inttype(distances), large_dist=large_dist,
            )
            kern(z, y, x, distances)
        else:
            kern = _get_aniso_distance_kernel(int_type=_get_inttype(distances))
            sampling = cupy.asarray(sampling, dtype=distances.dtype)
            kern(z, y, x, sampling, distances)
        vals = vals + (distances,)
    if return_indices:
        if indices_inplace:
            _check_indices(indices, (arr.ndim,) + arr.shape, x.dtype.itemsize)
            indices[0, ...] = z
            indices[1, ...] = y
            indices[2, ...] = x
        else:
            indices = cupy.stack((z, y, x), axis=0)
        vals = vals + (indices,)
    return vals



cdef object _thread_local
cdef dict _contexts


cdef class _OptimizationConfig:

    cdef readonly object optimize_impl
    cdef readonly int max_trials
    cdef readonly float timeout
    cdef readonly float expected_total_time_per_trial
    cdef readonly float max_total_time_per_trial


cdef class _OptimizationContext:

    cdef readonly str key
    cdef readonly _OptimizationConfig config
    cdef readonly dict _params_map
    cdef readonly bint _dirty


cpdef _OptimizationContext get_current_context()



from cupy import _util

# expose cache handles to this module
from cupy.fft._cache import get_plan_cache  # NOQA
from cupy.fft._cache import clear_plan_cache  # NOQA
from cupy.fft._cache import get_plan_cache_size  # NOQA
from cupy.fft._cache import set_plan_cache_size  # NOQA
from cupy.fft._cache import get_plan_cache_max_memsize  # NOQA
from cupy.fft._cache import set_plan_cache_max_memsize  # NOQA
from cupy.fft._cache import show_plan_cache_info  # NOQA

# on Linux, expose callback handles to this module
import sys as _sys
if _sys.platform.startswith('linux'):
    from cupy.fft._callback import get_current_callback_manager  # NOQA
    from cupy.fft._callback import set_cufft_callbacks  # NOQA
else:
    def get_current_callback_manager(*args, **kwargs):
        return None

    class set_cufft_callbacks:  # type: ignore
        def __init__(self, *args, **kwargs):
            raise RuntimeError('cuFFT callback is only available on Linux')


enable_nd_planning = True
use_multi_gpus = False
_devices = None


def set_cufft_gpus(gpus):
    '''Set the GPUs to be used in multi-GPU FFT.

    Args:
        gpus (int or list of int): The number of GPUs or a list of GPUs
            to be used. For the former case, the first ``gpus`` GPUs
            will be used.

    .. warning::
        This API is currently experimental and may be changed in the future
        version.

    .. seealso:: `Multiple GPU cuFFT Transforms`_

    .. _Multiple GPU cuFFT Transforms:
        https://docs.nvidia.com/cuda/cufft/index.html#multiple-GPU-cufft-transforms
    '''
    _util.experimental('cupy.fft.config.set_cufft_gpus')
    global _devices

    if isinstance(gpus, int):
        devs = [i for i in range(gpus)]
    elif isinstance(gpus, list):
        devs = gpus
    else:
        raise ValueError("gpus must be an int or a list of int.")
    if len(devs) <= 1:
        raise ValueError("Must use at least 2 GPUs.")

    # make it hashable
    _devices = tuple(devs)



# distutils: language = c++
import functools

import numpy

from cupy._core._kernel import ElementwiseKernel
from cupy._core._ufuncs import elementwise_copy
import cupy._core.core as core
from cupy.exceptions import AxisError

cimport cpython  # NOQA
cimport cython  # NOQA
from libcpp cimport vector

from cupy._core._dtype cimport get_dtype, _raise_if_invalid_cast
from cupy._core cimport core
from cupy._core.core cimport _ndarray_base
from cupy._core cimport internal
from cupy._core._kernel cimport _check_peer_access, _preprocess_args

from cupy.cuda import device


@cython.final
cdef class broadcast:
    """Object that performs broadcasting.

    CuPy actually uses this class to support broadcasting in various
    operations. Note that this class does not provide an iterator.

    Args:
        arrays (tuple of arrays): Arrays to be broadcasted.

    Attributes:
        ~broadcast.shape (tuple of ints): The broadcasted shape.
        nd (int): Number of dimensions of the broadcasted shape.
        ~broadcast.size (int): Total size of the broadcasted shape.
        values (list of arrays): The broadcasted arrays.

    .. seealso:: :class:`numpy.broadcast`

    """

    def __init__(self, *arrays):
        cdef shape_t shape
        cdef list val = list(arrays)
        internal._broadcast_core(val, shape)
        self.values = tuple(val)
        self.shape = tuple(shape)
        self.nd = <Py_ssize_t>shape.size()
        self.size = internal.prod(shape)


# _ndarray_base members


cdef _ndarray_shape_setter(_ndarray_base self, newshape):
    cdef shape_t shape, strides
    if not cpython.PySequence_Check(newshape):
        newshape = (newshape,)
    shape = internal.infer_unknown_dimension(newshape, self.size)
    _get_strides_for_nocopy_reshape(self, shape, strides)
    if strides.size() != shape.size():
        raise AttributeError(
            'Incompatible shape for in-place modification. Use `.reshape()` '
            'to make a copy with the desired shape.')
    self._set_shape_and_strides(shape, strides, False, True)


cdef _ndarray_base _ndarray_reshape(_ndarray_base self, tuple shape, order):
    cdef int order_char = internal._normalize_order(order, False)

    if len(shape) == 1 and cpython.PySequence_Check(shape[0]):
        shape = tuple(shape[0])

    if order_char == b'A':
        if self._f_contiguous and not self._c_contiguous:
            order_char = b'F'
        else:
            order_char = b'C'
    if order_char == b'C':
        return _reshape(self, shape)
    else:
        # TODO(grlee77): Support order within _reshape instead

        # The Fortran-ordered case is equivalent to:
        #     1.) reverse the axes via transpose
        #     2.) C-ordered reshape using reversed shape
        #     3.) reverse the axes via transpose
        return _T(_reshape(_T(self), shape[::-1]))


cdef _ndarray_base _ndarray_transpose(_ndarray_base self, tuple axes):
    if len(axes) == 0:
        return _T(self)
    if len(axes) == 1:
        a = axes[0]
        if a is None:
            return _T(self)
        elif cpython.PySequence_Check(a):
            axes = tuple(a)
    return _transpose(self, axes)


cdef _ndarray_base _ndarray_swapaxes(
        _ndarray_base self, Py_ssize_t axis1, Py_ssize_t axis2):
    cdef Py_ssize_t ndim = self.ndim
    cdef vector.vector[Py_ssize_t] axes
    if axis1 < -ndim or axis1 >= ndim or axis2 < -ndim or axis2 >= ndim:
        raise ValueError('Axis out of range')
    axis1 %= ndim
    axis2 %= ndim
    for i in range(ndim):
        axes.push_back(i)
    axes[axis1], axes[axis2] = axes[axis2], axes[axis1]
    return _transpose(self, axes)


cdef _ndarray_base _ndarray_flatten(_ndarray_base self, order):
    cdef int order_char
    cdef vector.vector[Py_ssize_t] axes

    order_char = internal._normalize_order(order, True)
    if order_char == b'A':
        if self._f_contiguous and not self._c_contiguous:
            order_char = b'F'
        else:
            order_char = b'C'
    if order_char == b'C':
        return _ndarray_flatten_order_c(self)
    elif order_char == b'F':
        return _ndarray_flatten_order_c(_T(self))
    elif order_char == b'K':
        axes = _npyiter_k_order_axes(self.strides)
        return _ndarray_flatten_order_c(_transpose(self, axes))


cdef _ndarray_base _ndarray_flatten_order_c(_ndarray_base self):
    newarray = self.copy(order='C')
    newarray._shape.assign(<Py_ssize_t>1, self.size)
    newarray._strides.assign(<Py_ssize_t>1,
                             <Py_ssize_t>self.itemsize)
    newarray._c_contiguous = True
    newarray._f_contiguous = True
    return newarray


cdef vector.vector[Py_ssize_t] _npyiter_k_order_axes(strides_t& strides):
    # output transpose axes such that
    # x.flatten(order="K") == x.transpose(axes).flatten(order="C")
    # by reproducing `npyiter_find_best_axis_ordering`
    # in numpy/core/src/multiarray/nditer_constr.c

    # Note that `flatten` and `ravel` should use this function for order="K",
    # while `copy(order="K")` should use `internal._get_strides_for_order_K`.
    cdef vector.vector[Py_ssize_t] axes
    cdef Py_ssize_t stride0, stride1
    cdef int ndim, i0, i1, ipos, k
    ndim = strides.size()
    for i0 in reversed(range(ndim)):
        stride0 = abs(strides[i0])
        if stride0 == 0:  # ambiguous
            axes.insert(axes.begin(), i0)
            continue
        ipos = 0
        for k, i1 in enumerate(axes):
            stride1 = abs(strides[i1])
            if stride1 == 0:  # ambiguous
                continue
            elif stride1 <= stride0:  # shouldswap = false
                break
            else:  # shouldswap = true
                ipos = k + 1
        axes.insert(axes.begin() + ipos, i0)
    return axes


cdef _ndarray_base _ndarray_ravel(_ndarray_base self, order):
    cdef int order_char
    cdef shape_t shape
    cdef vector.vector[Py_ssize_t] axes
    shape.push_back(self.size)

    order_char = internal._normalize_order(order, True)
    if order_char == b'A':
        if self._f_contiguous and not self._c_contiguous:
            order_char = b'F'
        else:
            order_char = b'C'
    if order_char == b'C':
        return _reshape(self, shape)
    elif order_char == b'F':
        return _reshape(_T(self), shape)
    elif order_char == b'K':
        axes = _npyiter_k_order_axes(self.strides)
        return _reshape(_transpose(self, axes), shape)


cdef _ndarray_base _ndarray_squeeze(_ndarray_base self, axis):
    cdef vector.vector[char] axis_flags
    cdef shape_t newshape
    cdef strides_t newstrides
    cdef Py_ssize_t ndim, naxes, _axis

    ndim = self._shape.size()
    axis_flags = vector.vector[char](ndim, 0)

    # Convert axis to boolean flag.
    if axis is None:
        for idim in range(ndim):
            if self._shape[idim] == 1:
                axis_flags[idim] = 1
    elif isinstance(axis, tuple):
        naxes = <Py_ssize_t>len(axis)
        for i in range(naxes):
            _axis = internal._normalize_axis_index(<Py_ssize_t>axis[i], ndim)
            if axis_flags[_axis] == 1:
                raise ValueError('duplicate value in \'axis\'')
            axis_flags[_axis] = 1
    else:
        _axis = <Py_ssize_t>axis
        if ndim == 0 and (_axis == 0 or _axis == -1):
            # Special case letting axis={-1,0} slip through for scalars,
            # for backwards compatibility reasons.
            pass
        else:
            _axis = internal._normalize_axis_index(_axis, ndim)
            axis_flags[_axis] = 1

    # Verify that the axes requested are all of size one
    any_ones = 0
    for idim in range(ndim):
        if axis_flags[idim] != 0:
            if self._shape[idim] == 1:
                any_ones = 1
            else:
                raise ValueError('cannot select an axis to squeeze out '
                                 'which has size not equal to one')

    # If there were no axes to squeeze out, return the same array
    if any_ones == 0:
        return self

    for i in range(ndim):
        if axis_flags[i] == 0:
            newshape.push_back(self._shape[i])
            newstrides.push_back(self._strides[i])

    v = self.view()
    # TODO(niboshi): Confirm update_x_contiguity flags
    v._set_shape_and_strides(newshape, newstrides, False, True)
    return v


cdef _ndarray_base _ndarray_repeat(_ndarray_base self, repeats, axis):
    return _repeat(self, repeats, axis)


# exposed


cpdef _ndarray_base _expand_dims(_ndarray_base a, tuple axis):
    cdef vector.vector[Py_ssize_t] normalized_axis
    cdef out_ndim = a.ndim + len(axis)
    cdef shape_t a_shape = a.shape, out_shape
    _normalize_axis_tuple(axis, out_ndim, normalized_axis)
    out_shape.assign(out_ndim, 0)
    cdef Py_ssize_t i, j
    for i in normalized_axis:
        out_shape[i] = 1
    j = 0
    for i in range(out_ndim):
        if out_shape[i] == 1:
            continue
        out_shape[i] = a_shape[j]
        j += 1
    return _reshape(a, out_shape)


cpdef _ndarray_base moveaxis(_ndarray_base a, source, destination):
    cdef shape_t src, dest
    cdef Py_ssize_t ndim = a.ndim
    _normalize_axis_tuple(source, ndim, src)
    _normalize_axis_tuple(destination, ndim, dest)

    if src.size() != dest.size():
        raise ValueError('`source` and `destination` arguments must have '
                         'the same number of elements')

    cdef vector.vector[Py_ssize_t] order
    cdef Py_ssize_t i
    for i in range(ndim):
        if not _has_element(src, i):
            order.push_back(i)

    cdef Py_ssize_t d, s
    for d, s in sorted(zip(dest, src)):
        order.insert(order.begin() + d, s)

    return _transpose(a, order)


cpdef _ndarray_base _move_single_axis(
        _ndarray_base a, Py_ssize_t source, Py_ssize_t destination):
    """Like moveaxis, but supporting only integer source and destination."""
    cdef Py_ssize_t ndim = a.ndim
    source = internal._normalize_axis_index(source, ndim)
    destination = internal._normalize_axis_index(destination, ndim)

    if source == destination:
        return a

    cdef vector.vector[Py_ssize_t] order
    cdef Py_ssize_t i
    for i in range(ndim):
        if i != source:
            order.push_back(i)

    order.insert(order.begin() + destination, source)
    return _transpose(a, order)


cpdef _ndarray_base rollaxis(
        _ndarray_base a, Py_ssize_t axis, Py_ssize_t start=0):
    cdef Py_ssize_t i, ndim = a.ndim
    cdef vector.vector[Py_ssize_t] axes
    if axis < 0:
        axis += ndim
    if start < 0:
        start += ndim
    if not (0 <= axis < ndim and 0 <= start <= ndim):
        raise ValueError('Axis out of range')
    if axis < start:
        start -= 1
    if axis == start:
        return a
    if ndim == 2:
        return _transpose(a, axes)

    for i in range(ndim):
        axes.push_back(i)
    axes.erase(axes.begin() + axis)
    axes.insert(axes.begin() + start, axis)
    return _transpose(a, axes)


cpdef _ndarray_base _reshape(_ndarray_base self, const shape_t &shape_spec):
    cdef shape_t shape
    cdef strides_t strides
    cdef _ndarray_base newarray
    shape = internal.infer_unknown_dimension(shape_spec, self.size)
    if internal.vector_equal(shape, self._shape):
        return self.view()

    _get_strides_for_nocopy_reshape(self, shape, strides)
    if strides.size() == shape.size():
        return self._view(type(self), shape, strides, False, True, self)
    newarray = self.copy()
    _get_strides_for_nocopy_reshape(newarray, shape, strides)

    # TODO(niboshi): Confirm update_x_contiguity flags
    newarray._set_shape_and_strides(shape, strides, False, True)
    return newarray


cpdef _ndarray_base _T(_ndarray_base self):
    ret = self.view()
    ret._shape.assign(self._shape.rbegin(), self._shape.rend())
    ret._strides.assign(self._strides.rbegin(), self._strides.rend())
    ret._c_contiguous = self._f_contiguous
    ret._f_contiguous = self._c_contiguous
    return ret


cpdef _ndarray_base _transpose(
        _ndarray_base self, const vector.vector[Py_ssize_t] &axes):
    cdef vector.vector[Py_ssize_t] a_axes
    cdef vector.vector[char] axis_flags
    cdef Py_ssize_t i, ndim, axis, axes_size
    cdef bint is_normal = True, is_trans = True

    axes_size = axes.size()
    if axes_size == 0:
        return _T(self)

    ndim = self._shape.size()
    if axes_size != ndim:
        raise ValueError("axes don't match array")

    axis_flags.resize(ndim, 0)
    for i in range(axes_size):
        axis = axes[i]
        if axis < -ndim or axis >= ndim:
            raise AxisError(axis, ndim)
        axis %= ndim
        a_axes.push_back(axis)
        if axis_flags[axis]:
            raise ValueError('repeated axis in transpose')
        axis_flags[axis] = 1
        is_normal &= i == axis
        is_trans &= ndim - 1 - i == axis

    if is_normal:
        return self.view()
    if is_trans:
        return _T(self)

    ret = self.view()
    ret._shape.clear()
    ret._strides.clear()
    for axis in a_axes:
        ret._shape.push_back(self._shape[axis])
        ret._strides.push_back(self._strides[axis])
    ret._update_contiguity()
    return ret


cpdef array_split(_ndarray_base ary, indices_or_sections, Py_ssize_t axis):
    cdef Py_ssize_t i, ndim, size, each_size, index, prev, stride
    cdef Py_ssize_t num_large
    cdef shape_t shape

    ndim = ary.ndim
    if -ndim > axis or ndim <= axis:
        raise IndexError('Axis exceeds ndim')
    if axis < 0:
        axis += ndim
    size = ary._shape[axis]

    if numpy.isscalar(indices_or_sections):
        each_size = (size - 1) // indices_or_sections
        num_large = (size - 1) % indices_or_sections + 1
        indices = [i * each_size + min(i, num_large)
                   for i in range(1, indices_or_sections)]
    else:
        indices = [i if i >= 0 else size + i for i in indices_or_sections]

    if len(indices) == 0:
        return [ary]

    # Make a copy of shape for each view
    shape = ary._shape

    prev = 0
    ret = []
    stride = ary._strides[axis]
    if ary.size == 0:
        stride = 0
    for index in indices:
        index = min(index, size)
        shape[axis] = max(index - prev, 0)
        v = ary.view()
        v.data = ary.data + prev * stride
        # TODO(niboshi): Confirm update_x_contiguity flags
        v._set_shape_and_strides(shape, ary._strides, True, True)
        ret.append(v)

        prev = index

    shape[axis] = size - prev
    v = ary.view()
    v.data = ary.data + prev * stride
    # TODO(niboshi): Confirm update_x_contiguity flags
    v._set_shape_and_strides(shape, ary._strides, True, True)
    ret.append(v)

    return ret


cpdef _ndarray_base broadcast_to(_ndarray_base array, shape):
    """Broadcast an array to a given shape.

    .. seealso::
        :func:`cupy.broadcast_to` for full documentation,
        :meth:`numpy.broadcast_to`

    """
    shape = tuple(shape) if numpy.iterable(shape) else (shape,)
    cdef int i, j, ndim = array._shape.size(), length = len(shape)
    cdef Py_ssize_t sh, a_sh
    if ndim > length:
        raise ValueError(
            'input operand has more dimensions than allowed by the axis '
            'remapping')
    cdef shape_t _shape = shape
    cdef strides_t strides
    strides.assign(length, 0)
    for i in range(ndim):
        j = i + length - ndim
        sh = _shape[j]
        a_sh = array._shape[i]
        if sh == a_sh:
            strides[j] = array._strides[i]
        elif a_sh != 1:
            raise ValueError(
                'operands could not be broadcast together with shape {} and '
                'requested shape {}'.format(array.shape, shape))

    view = array.view()
    # TODO(niboshi): Confirm update_x_contiguity flags
    view._set_shape_and_strides(_shape, strides, True, True)
    return view


cpdef _ndarray_base _repeat(_ndarray_base a, repeats, axis=None):
    """Repeat arrays along an axis.

    Args:
        a (cupy.ndarray): Array to transform.
        repeats (int, list or tuple): The number of repeats.
        axis (int): The axis to repeat.

    Returns:
        cupy.ndarray: Transformed array with repeats.

    .. seealso:: :func:`numpy.repeat`

    """
    cdef _ndarray_base ret

    if isinstance(repeats, _ndarray_base):
        raise ValueError(
            'cupy.ndaray cannot be specified as `repeats` argument.')

    # Scalar and size 1 'repeat' arrays broadcast to any shape, for all
    # other inputs the dimension must match exactly.
    cdef bint broadcast = False
    # numpy.issubdtype(1, numpy.integer) fails with old numpy like 1.13.3.
    if (isinstance(repeats, int) or
            (hasattr(repeats, 'dtype') and
             numpy.issubdtype(repeats, numpy.integer))):
        if repeats < 0:
            raise ValueError(
                '\'repeats\' should not be negative: {}'.format(repeats))
        broadcast = True
        repeats = [repeats]
    elif cpython.PySequence_Check(repeats):
        for rep in repeats:
            if rep < 0:
                raise ValueError(
                    'all elements of \'repeats\' should not be negative: {}'
                    .format(repeats))
        if len(repeats) == 1:
            broadcast = True
    else:
        raise ValueError(
            '\'repeats\' should be int or sequence: {}'.format(repeats))

    if axis is None:
        if broadcast:
            a = _reshape(a, (-1, 1))
            ret = core.ndarray((a.size, repeats[0]), dtype=a.dtype)
            if ret.size:
                elementwise_copy(a, ret)
            return ret.ravel()
        else:
            a = a.ravel()
            axis = 0
    else:
        axis = internal._normalize_axis_index(axis, a.ndim)

    if broadcast:
        repeats = repeats * a._shape[axis]
    elif a.shape[axis] != len(repeats):
        raise ValueError(
            '\'repeats\' and \'axis\' of \'a\' should be same length: {} != {}'
            .format(a.shape[axis], len(repeats)))

    ret_shape = list(a.shape)
    ret_shape[axis] = sum(repeats)
    ret = core.ndarray(ret_shape, dtype=a.dtype)
    a_index = [slice(None)] * len(ret_shape)
    ret_index = list(a_index)
    offset = 0
    for i in range(a._shape[axis]):
        if repeats[i] == 0:
            continue
        a_index[axis] = slice(i, i + 1)
        ret_index[axis] = slice(offset, offset + repeats[i])
        # convert to tuple because cupy has a indexing bug
        ret[tuple(ret_index)] = a[tuple(a_index)]
        offset += repeats[i]
    return ret


cpdef _ndarray_base concatenate_method(
        tup, int axis, _ndarray_base out=None, dtype=None,
        casting='same_kind'):
    cdef int ndim0
    cdef int i
    cdef _ndarray_base a, a0

    if dtype is not None:
        dtype = get_dtype(dtype)

    dev_id = device.get_device_id()
    arrays = _preprocess_args(dev_id, tup, False)

    # Check if the input is not an empty sequence
    if len(arrays) == 0:
        raise ValueError('Cannot concatenate from empty tuple')

    # Check types of the input arrays
    for o in arrays:
        if not isinstance(o, _ndarray_base):
            raise TypeError('Only cupy arrays can be concatenated')

    # Check ndim > 0 for the input arrays
    for o in arrays:
        a = o
        if a._shape.size() == 0:
            raise TypeError('zero-dimensional arrays cannot be concatenated')

    # Check ndim consistency of the input arrays
    a0 = arrays[0]
    ndim0 = a0._shape.size()
    for o in arrays[1:]:
        a = o
        if <int>(a._shape.size()) != ndim0:
            raise ValueError(
                'All arrays to concatenate must have the same ndim')

    # Check shape consistency of the input arrays, and compute the output shape
    shape0 = a0._shape
    axis = internal._normalize_axis_index(axis, ndim0)
    for o in arrays[1:]:
        a = o
        for i in range(ndim0):
            if i != axis and shape0[i] != a._shape[i]:
                raise ValueError(
                    'All arrays must have same shape except the axis to '
                    'concatenate')
        shape0[axis] += a._shape[axis]

    # Compute the output dtype
    if out is None:
        if dtype is None:
            dtype = a0.dtype
            have_same_types = True
            for o in arrays[1:]:
                have_same_types = have_same_types and (o.dtype == dtype)
            if not have_same_types:
                dtype = functools.reduce(
                    numpy.promote_types, set([a.dtype for a in arrays]))
    else:
        if dtype is not None:
            raise TypeError('concatenate() only takes `out` or `dtype` as an '
                            'argument, but both were provided.')
        dtype = out.dtype

    # Check casting rule
    for o in arrays:
        _raise_if_invalid_cast(o.dtype, dtype, casting)

    # Prpare the output array
    shape_t = tuple(shape0)
    if out is None:
        out = core.ndarray(shape_t, dtype=dtype)
    else:
        if len(out.shape) != len(shape_t):
            raise ValueError('Output array has wrong dimensionality')
        if out.shape != shape_t:
            raise ValueError('Output array is the wrong shape')

    return _concatenate(arrays, axis, shape_t, out, casting)


cpdef _ndarray_base _concatenate(
        list arrays, Py_ssize_t axis, tuple shape, _ndarray_base out,
        str casting):
    cdef _ndarray_base a, b
    cdef Py_ssize_t i, aw, itemsize, axis_size
    cdef bint all_same_type, same_shape_and_contiguous
    # If arrays are large, Issuing each copy method is efficient.
    cdef Py_ssize_t threshold_size = 2 * 1024 * 1024

    dtype = out.dtype

    if len(arrays) > 8:
        all_same_type = True
        same_shape_and_contiguous = True
        axis_size = shape[axis] // len(arrays)
        total_bytes = 0
        itemsize = dtype.itemsize
        for a in arrays:
            if a.dtype != dtype:
                all_same_type = False
                break
            if same_shape_and_contiguous:
                same_shape_and_contiguous = (
                    a._c_contiguous and a._shape[axis] == axis_size)
            total_bytes += a.size * itemsize

        if all_same_type and total_bytes < threshold_size * len(arrays):
            return _concatenate_single_kernel(
                arrays, axis, shape, dtype, same_shape_and_contiguous, out)

    i = 0
    slice_list = [slice(None)] * len(shape)
    for a in arrays:
        aw = a._shape[axis]
        slice_list[axis] = slice(i, i + aw)
        b = out[tuple(slice_list)]
        elementwise_copy(a, b, casting=casting)
        i += aw
    return out


cpdef Py_ssize_t size(_ndarray_base a, axis=None) except? -1:
    """Returns the number of elements along a given axis.

    Args:
        a (ndarray): Input data.
        axis (int or None): Axis along which the elements are counted.
            When it is ``None``, it returns the total number of elements.

    Returns:
        int: Number of elements along the given axis.

    """
    cdef int index, ndim
    if axis is None:
        return a.size
    else:
        index = axis
        ndim = a._shape.size()
        if index < 0:
            index += ndim
        if not 0 <= index < ndim:
            raise IndexError('index out of range')
        return a._shape[index]


# private


cdef bint _has_element(const shape_t &source, Py_ssize_t n):
    for i in range(source.size()):
        if source[i] == n:
            return True
    return False


cdef _get_strides_for_nocopy_reshape(
        _ndarray_base a, const shape_t &newshape, strides_t &newstrides):
    cdef Py_ssize_t size, itemsize, ndim, dim, last_stride
    size = a.size
    newstrides.clear()

    itemsize = a.itemsize
    if size == 1:
        newstrides.assign(<Py_ssize_t>newshape.size(), itemsize)
        return
    if size == 0:
        internal.get_contiguous_strides_inplace(
            newshape, newstrides, itemsize, True, False)
        return

    cdef shape_t shape
    cdef strides_t strides
    internal.get_reduced_dims(a._shape, a._strides, itemsize, shape, strides)

    ndim = shape.size()
    dim = 0
    last_stride = shape[0] * strides[0]
    for i in range(newshape.size()):
        size = newshape[i]
        if size <= 1:
            newstrides.push_back(last_stride)
            continue
        if dim >= ndim or shape[dim] % size != 0:
            newstrides.clear()
            break
        shape[dim] //= size
        last_stride = shape[dim] * strides[dim]
        newstrides.push_back(last_stride)
        if shape[dim] == 1:
            dim += 1


cdef _normalize_axis_tuple(axis, Py_ssize_t ndim, shape_t &ret):
    """Normalizes an axis argument into a tuple of non-negative integer axes.

    Arguments `argname` and `allow_duplicate` are not supported.

    """
    if numpy.isscalar(axis):
        axis = (axis,)

    for ax in axis:
        ax = internal._normalize_axis_index(ax, ndim)
        if _has_element(ret, ax):
            # the message in `numpy.core.numeric.normalize_axis_tuple`
            raise ValueError('repeated axis')
        ret.push_back(ax)


cdef _ndarray_base _concatenate_single_kernel(
        list arrays, Py_ssize_t axis, tuple shape, dtype,
        bint same_shape_and_contiguous, _ndarray_base out):
    cdef _ndarray_base a, x
    cdef Py_ssize_t base, cum, ndim
    cdef int i, j
    cdef Py_ssize_t[:] ptrs
    cdef Py_ssize_t[:] cum_sizes
    cdef Py_ssize_t[:, :] x_strides
    cdef int device_id = device.get_device_id()

    assert out is not None

    ptrs = numpy.ndarray(len(arrays), numpy.int64)
    for i, a in enumerate(arrays):
        _check_peer_access(a, device_id)
        ptrs[i] = a.data.ptr
    x = core.array(ptrs)

    if same_shape_and_contiguous:
        base = internal.prod_sequence(shape[axis:]) // len(arrays)
        _concatenate_kernel_same_size(x, base, out)
        return out

    ndim = len(shape)
    x_strides = numpy.ndarray((len(arrays), ndim), numpy.int64)
    cum_sizes = numpy.ndarray(len(arrays), numpy.int64)
    cum = 0
    for i, a in enumerate(arrays):
        for j in range(ndim):
            x_strides[i, j] = <int>a._strides[j]
        cum_sizes[i] = cum
        cum += <int>a._shape[axis]

    _concatenate_kernel(
        x, axis, core.array(cum_sizes), core.array(x_strides), out)
    return out


cdef _concatenate_kernel_same_size = ElementwiseKernel(
    'raw P x, int64 base',
    'T y',
    '''
    ptrdiff_t middle = i / base;
    ptrdiff_t top = middle / x.size();
    ptrdiff_t array_ind = middle - top * x.size();
    ptrdiff_t offset = i + (top - middle) * base;
    y = reinterpret_cast<T*>(x[array_ind])[offset];
    ''',
    'cupy_concatenate_same_size'
)


cdef _concatenate_kernel = ElementwiseKernel(
    '''raw P x, int32 axis, raw int64 cum_sizes, raw int64 x_strides''',
    'T y',
    '''
    ptrdiff_t axis_ind = _ind.get()[axis];
    ptrdiff_t left = 0;
    ptrdiff_t right = cum_sizes.size();

    while (left < right - 1) {
      ptrdiff_t m = (left + right) / 2;
      if (axis_ind < cum_sizes[m]) {
        right = m;
      } else {
        left = m;
      }
    }

    ptrdiff_t array_ind = left;
    axis_ind -= cum_sizes[left];
    char* ptr = reinterpret_cast<char*>(x[array_ind]);
    for (int j = _ind.ndim - 1; j >= 0; --j) {
      ptrdiff_t ind[] = {array_ind, j};
      ptrdiff_t offset;
      if (j == axis) {
        offset = axis_ind;
      } else {
        offset = _ind.get()[j];
      }
      ptr += x_strides[ind] * offset;
    }

    y = *reinterpret_cast<T*>(ptr);
    ''',
    'cupy_concatenate',
    reduce_dims=False
)



import math
import numbers
import os

import cupy

from ._util import _get_inttype

if hasattr(math, 'lcm'):
    lcm = math.lcm
else:
    """Fallback implementation of least common multiple (lcm)"""
    def _lcm(a, b):
        return abs(b * (a // math.gcd(a, b)))

    def lcm(*integers):
        nargs = len(integers)
        if not all(isinstance(a, numbers.Integral) for a in integers):
            raise TypeError("all arguments must be integers")
        if nargs == 0:
            return 1
        res = int(integers[0])
        if nargs == 1:
            return abs(res)
        for i in range(1, nargs):
            x = int(integers[i])
            res = _lcm(res, x)
        return res


pba2d_defines_template = """

// MARKER is used to mark blank pixels in the texture.
// Any uncolored pixels will have x = MARKER.
// Input texture should have x = MARKER for all pixels other than sites
#define MARKER      {marker}
#define BLOCKSIZE   {block_size_2d}
#define pixel_int2_t {pixel_int2_t}                // typically short2 (int2 for images with > 32k pixels per side)
#define make_pixel(x, y)  {make_pixel_func}(x, y)  // typically make_short2 (make_int2 images with > 32k pixels per side

"""  # noqa


def _init_marker(int_dtype):
    """use a minimum value that is appropriate to the integer dtype"""
    if int_dtype == cupy.int16:
        # marker = cupy.iinfo(int_dtype).min
        marker = -32768
    elif int_dtype == cupy.int32:
        # divide by two so we don't have to promote other intermediate int
        # variables to 64-bit int
        marker = -2147483648 // 2
    else:
        raise ValueError(
            "expected int_dtype to be either cupy.int16 or cupy.int32"
        )
    return marker


@cupy.memoize(True)
def get_pba2d_src(block_size_2d=64, marker=-32768, pixel_int2_t="short2"):
    make_pixel_func = "make_" + pixel_int2_t

    pba2d_code = pba2d_defines_template.format(
        block_size_2d=block_size_2d,
        marker=marker,
        pixel_int2_t=pixel_int2_t,
        make_pixel_func=make_pixel_func
    )
    kernel_directory = os.path.join(os.path.dirname(__file__), "cuda")
    with open(os.path.join(kernel_directory, "pba_kernels_2d.h"), "rt") as f:
        pba2d_kernels = "\n".join(f.readlines())

    pba2d_code += pba2d_kernels
    return pba2d_code


def _get_block_size(check_warp_size=False):
    if check_warp_size:
        dev = cupy.cuda.runtime.getDevice()
        device_properties = cupy.cuda.runtime.getDeviceProperties(dev)
        return int(device_properties["warpSize"])
    else:
        return 32


@cupy.memoize(for_each_device=True)
def _get_pack_kernel(int_type, marker=-32768):
    """Pack coordinates into array of type short2 (or int2).

    This kernel works with 2D input data, `arr` (typically boolean).

    The output array, `out` will be 3D with a signed integer dtype.
    It will have size 2 on the last axis so that it can be viewed as a CUDA
    vector type such as `int2` or `float2`.
    """
    code = f"""
    if (arr[i]) {{
        out[2*i] = {marker};
        out[2*i + 1] = {marker};
    }} else {{
        int shape_1 = arr.shape()[1];
        int _i = i;
        int ind_1 = _i % shape_1;
        _i /= shape_1;
        out[2*i] = ind_1;   // out.x
        out[2*i + 1] = _i;  // out.y
    }}
    """
    return cupy.ElementwiseKernel(
        in_params="raw B arr",
        out_params="raw I out",
        operation=code,
        options=("--std=c++11",),
    )


def _pack_int2(arr, marker=-32768, int_dtype=cupy.int16):
    if arr.ndim != 2:
        raise ValueError("only 2d arr supported")
    int2_dtype = cupy.dtype({"names": ["x", "y"], "formats": [int_dtype] * 2})
    out = cupy.zeros(arr.shape + (2,), dtype=int_dtype)
    assert out.size == 2 * arr.size
    pack_kernel = _get_pack_kernel(
        int_type="short" if int_dtype == cupy.int16 else "int",
        marker=marker
    )
    pack_kernel(arr, out, size=arr.size)
    out = cupy.squeeze(out.view(int2_dtype))
    return out


def _unpack_int2(img, make_copy=False, int_dtype=cupy.int16):
    temp = img.view(int_dtype).reshape(img.shape + (2,))
    if make_copy:
        temp = temp.copy()
    return temp


def _determine_padding(shape, padded_size, block_size):
    # all kernels assume equal size along both axes, so pad up to equal size if
    # shape is not isotropic
    orig_sy, orig_sx = shape
    if orig_sx != padded_size or orig_sy != padded_size:
        padding_width = (
            (0, padded_size - orig_sy), (0, padded_size - orig_sx)
        )
    else:
        padding_width = None
    return padding_width


def _generate_shape(ndim, int_type, var_name="out", raw_var=True):
    code = ""
    if not raw_var:
        var_name = "_raw_" + var_name
    for i in range(ndim):
        code += f"{int_type} shape_{i} = {var_name}.shape()[{i}];\n"
    return code


def _generate_indices_ops(ndim, int_type):
    code = f"{int_type} _i = i;\n"
    for j in range(ndim - 1, 0, -1):
        code += f"{int_type} ind_{j} = _i % shape_{j};\n_i /= shape_{j};\n"
    code += f"{int_type} ind_0 = _i;"
    return code


def _get_distance_kernel_code(int_type, dist_int_type, raw_out_var=True):
    code = _generate_shape(
        ndim=2, int_type=int_type, var_name="dist", raw_var=raw_out_var
    )
    code += _generate_indices_ops(ndim=2, int_type=int_type)
    code += f"""
    {int_type} tmp;
    {dist_int_type} sq_dist;
    tmp = y[i] - ind_0;
    sq_dist = tmp * tmp;
    tmp = x[i] - ind_1;
    sq_dist += tmp * tmp;
    dist[i] = sqrt(static_cast<F>(sq_dist));
    """
    return code


@cupy.memoize(for_each_device=True)
def _get_distance_kernel(int_type, dist_int_type):
    """Returns kernel computing the Euclidean distance from coordinates."""
    operation = _get_distance_kernel_code(
        int_type, dist_int_type, raw_out_var=True
    )
    return cupy.ElementwiseKernel(
        in_params="raw I y, raw I x",
        out_params="raw F dist",
        operation=operation,
        options=("--std=c++11",),
    )


def _get_aniso_distance_kernel_code(int_type, raw_out_var=True):
    code = _generate_shape(
        ndim=2, int_type=int_type, var_name="dist", raw_var=raw_out_var
    )
    code += _generate_indices_ops(ndim=2, int_type=int_type)
    code += """
    F tmp;
    F sq_dist;
    tmp = static_cast<F>(y[i] - ind_0) * sampling[0];
    sq_dist = tmp * tmp;
    tmp = static_cast<F>(x[i] - ind_1) * sampling[1];
    sq_dist += tmp * tmp;
    dist[i] = sqrt(sq_dist);
    """
    return code


@cupy.memoize(for_each_device=True)
def _get_aniso_distance_kernel(int_type):
    """Returns kernel computing the Euclidean distance from coordinates."""
    operation = _get_aniso_distance_kernel_code(int_type, raw_out_var=True)
    return cupy.ElementwiseKernel(
        in_params="raw I y, raw I x, raw F sampling",
        out_params="raw F dist",
        operation=operation,
        options=("--std=c++11",),
    )


def _distance_tranform_arg_check(distances_out, indices_out,
                                 return_distances, return_indices):
    """Raise a RuntimeError if the arguments are invalid"""
    error_msgs = []
    if (not return_distances) and (not return_indices):
        error_msgs.append(
            "at least one of return_distances/return_indices must be True")
    if distances_out and not return_distances:
        error_msgs.append(
            "return_distances must be True if distances is supplied"
        )
    if indices_out and not return_indices:
        error_msgs.append("return_indices must be True if indices is supplied")
    if error_msgs:
        raise RuntimeError(", ".join(error_msgs))


def _check_distances(distances, shape, dtype):
    if distances.shape != shape:
        raise RuntimeError("distances array has wrong shape")
    if distances.dtype != dtype:
        raise RuntimeError(
            f"distances array must have dtype: {dtype}")


def _check_indices(indices, shape, itemsize):
    if indices.shape != shape:
        raise RuntimeError("indices array has wrong shape")
    if indices.dtype.kind not in 'iu':
        raise RuntimeError(
            "indices array must have an integer dtype"
        )
    elif indices.dtype.itemsize < itemsize:
        raise RuntimeError(
            f"indices dtype must have itemsize > {itemsize}"
        )


def _pba_2d(arr, sampling=None, return_distances=True, return_indices=False,
            block_params=None, check_warp_size=False, *,
            float64_distances=False, distances=None, indices=None):

    indices_inplace = isinstance(indices, cupy.ndarray)
    dt_inplace = isinstance(distances, cupy.ndarray)
    _distance_tranform_arg_check(
        dt_inplace, indices_inplace, return_distances, return_indices
    )

    # input_arr: a 2D image
    #    For each site at (x, y), the pixel at coordinate (x, y) should contain
    #    the pair (x, y). Pixels that are not sites should contain the pair
    #    (MARKER, MARKER)

    # Note: could query warp size here, but for now just assume 32 to avoid
    #       overhead of querying properties
    block_size = _get_block_size(check_warp_size)

    if block_params is None:
        padded_size = math.ceil(max(arr.shape) / block_size) * block_size

        # should be <= size / block_size. sy must be a multiple of m1
        m1 = padded_size // block_size
        # size must be a multiple of m2
        m2 = max(1, min(padded_size // block_size, block_size))
        # m2 must also be a power of two
        m2 = 2**math.floor(math.log2(m2))
        if padded_size % m2 != 0:
            raise RuntimeError("error in setting default m2")
        m3 = min(min(m1, m2), 2)
    else:
        if any(p < 1 for p in block_params):
            raise ValueError("(m1, m2, m3) in blockparams must be >= 1")
        m1, m2, m3 = block_params
        if math.log2(m2) % 1 > 1e-5:
            raise ValueError("m2 must be a power of 2")
        multiple = lcm(block_size, m1, m2, m3)
        padded_size = math.ceil(max(arr.shape) / multiple) * multiple

    if m1 > padded_size // block_size:
        raise ValueError(
            f"m1 too large. must be <= padded arr.shape[0] // {block_size}"
        )
    if m2 > padded_size // block_size:
        raise ValueError(
            f"m2 too large. must be <= padded arr.shape[1] // {block_size}"
        )
    if m3 > padded_size // block_size:
        raise ValueError(
            f"m3 too large. must be <= padded arr.shape[1] // {block_size}"
        )
    for m in (m1, m2, m3):
        if padded_size % m != 0:
            raise ValueError(
                f"Largest dimension of image ({padded_size}) must be evenly "
                f"disivible by each element of block_params: {(m1, m2, m3)}."
            )

    shape_max = max(arr.shape)
    if shape_max <= 32768:
        int_dtype = cupy.int16
        pixel_int2_type = "short2"
    else:
        if shape_max > (1 << 24):
            # limit to coordinate range to 2**24 due to use of __mul24 in
            # coordinate TOID macro
            raise ValueError(
                f"maximum axis size of {1 << 24} exceeded, for image with "
                f"shape {arr.shape}"
            )
        int_dtype = cupy.int32
        pixel_int2_type = "int2"

    marker = _init_marker(int_dtype)

    orig_sy, orig_sx = arr.shape
    padding_width = _determine_padding(arr.shape, padded_size, block_size)
    if padding_width is not None:
        arr = cupy.pad(arr, padding_width, mode="constant", constant_values=1)
    size = arr.shape[0]

    input_arr = _pack_int2(arr, marker=marker, int_dtype=int_dtype)
    output = cupy.zeros_like(input_arr)

    int2_dtype = cupy.dtype({"names": ["x", "y"], "formats": [int_dtype] * 2})
    margin = cupy.empty((2 * m1 * size,), dtype=int2_dtype)

    # phase 1 of PBA. m1 must divide texture size and be <= 64
    pba2d = cupy.RawModule(
        code=get_pba2d_src(
            block_size_2d=block_size,
            marker=marker,
            pixel_int2_t=pixel_int2_type,
        )
    )
    kernelFloodDown = pba2d.get_function("kernelFloodDown")
    kernelFloodUp = pba2d.get_function("kernelFloodUp")
    kernelPropagateInterband = pba2d.get_function("kernelPropagateInterband")
    kernelUpdateVertical = pba2d.get_function("kernelUpdateVertical")
    kernelCreateForwardPointers = pba2d.get_function(
        "kernelCreateForwardPointers"
    )
    kernelDoubleToSingleList = pba2d.get_function("kernelDoubleToSingleList")

    if sampling is None:
        kernelProximatePoints = pba2d.get_function("kernelProximatePoints")
        kernelMergeBands = pba2d.get_function("kernelMergeBands")
        kernelColor = pba2d.get_function("kernelColor")
    else:
        kernelProximatePoints = pba2d.get_function(
            "kernelProximatePointsWithSpacing"
        )
        kernelMergeBands = pba2d.get_function("kernelMergeBandsWithSpacing")
        kernelColor = pba2d.get_function("kernelColorWithSpacing")

    block = (block_size, 1, 1)
    grid = (math.ceil(size / block[0]), m1, 1)
    bandSize1 = size // m1
    # kernelFloodDown modifies input_arr in-place
    kernelFloodDown(
        grid,
        block,
        (input_arr, input_arr, size, bandSize1),
    )
    # kernelFloodUp modifies input_arr in-place
    kernelFloodUp(
        grid,
        block,
        (input_arr, input_arr, size, bandSize1),
    )
    # kernelFloodUp fills values into margin
    kernelPropagateInterband(
        grid,
        block,
        (input_arr, margin, size, bandSize1),
    )
    # kernelUpdateVertical stores output into an intermediate array of
    # transposed shape
    kernelUpdateVertical(
        grid,
        block,
        (input_arr, margin, output, size, bandSize1),
    )

    # phase 2
    block = (block_size, 1, 1)
    grid = (math.ceil(size / block[0]), m2, 1)
    bandSize2 = size // m2
    if sampling is None:
        sampling_args = ()
    else:
        # Originally the shape is (y, x) and sampling[1] corresponds to y.
        # However, kernelUpdateVertical transposed the image, so
        # we are now working with (x, y) instead. Need sampling ordered
        # accordingly.
        sampling = tuple(map(float, sampling))
        sampling_args = (sampling[0], sampling[1])
    kernelProximatePoints(
        grid,
        block,
        (output, input_arr, size, bandSize2) + sampling_args,
    )
    kernelCreateForwardPointers(
        grid,
        block,
        (input_arr, input_arr, size, bandSize2),
    )
    # Repeatedly merging two bands into one
    noBand = m2
    while noBand > 1:
        grid = (math.ceil(size / block[0]), noBand // 2)
        kernelMergeBands(
            grid,
            block,
            (output, input_arr, input_arr, size, size // noBand) + sampling_args,  # noqa
        )
        noBand //= 2
    # Replace the forward link with the X coordinate of the seed to remove
    # the need of looking at the other texture. We need it for coloring.
    grid = (math.ceil(size / block[0]), size)
    kernelDoubleToSingleList(
        grid,
        block,
        (output, input_arr, input_arr, size),
    )

    # Phase 3 of PBA
    block = (block_size, m3, 1)
    grid = (math.ceil(size / block[0]), 1, 1)
    kernelColor(
        grid,
        block,
        (input_arr, output, size) + sampling_args,
    )

    output = _unpack_int2(output, make_copy=False, int_dtype=int_dtype)
    # make sure to crop any padding that was added here!
    x = output[:orig_sy, :orig_sx, 0]
    y = output[:orig_sy, :orig_sx, 1]

    vals = ()
    if return_distances:
        dtype_out = cupy.float64 if float64_distances else cupy.float32
        if dt_inplace:
            _check_distances(distances, y.shape, dtype_out)
        else:
            distances = cupy.zeros(y.shape, dtype=dtype_out)

        # make sure maximum possible distance doesn"t overflow
        max_possible_dist = sum((s - 1)**2 for s in y.shape)
        dist_int_type = "int" if max_possible_dist < 2**31 else "ptrdiff_t"

        if sampling is None:
            distance_kernel = _get_distance_kernel(
                int_type=_get_inttype(distances),
                dist_int_type=dist_int_type,
            )
            distance_kernel(y, x, distances, size=distances.size)
        else:
            distance_kernel = _get_aniso_distance_kernel(
                int_type=_get_inttype(distances),
            )
            sampling = cupy.asarray(sampling, dtype=dtype_out)
            distance_kernel(y, x, sampling, distances, size=distances.size)

        vals = vals + (distances,)
    if return_indices:
        if indices_inplace:
            _check_indices(indices, (arr.ndim,) + arr.shape, x.dtype.itemsize)
            indices[0, ...] = y
            indices[1, ...] = x
        else:
            indices = cupy.stack((y, x), axis=0)
        vals = vals + (indices,)
    return vals



import dataclasses
import typing
from typing import Callable, Optional

import cupy
from cupyx.distributed.array import _array
from cupyx.distributed.array import _chunk
from cupyx.distributed.array import _modes


_SliceIndices = tuple[int, int, int]


_BlockIdx = tuple[_SliceIndices, _SliceIndices]


# Where the chunks_map holds a specific block
# location_map[block_idx] = {dev: i if chunk_map[dev][i].index == block_idx}
_BlockLocationMap = dict[_BlockIdx, dict[int, int]]


# (block_a, block_b, device): device can compute block_a @ block_b
_ExecutionPlan = list[tuple[_BlockIdx, _BlockIdx, int]]


# Index within a batch pointing for a 2D matrix
# len(batch_idx) == ndim - 2
_BatchIdx = tuple[_SliceIndices, ...]


@dataclasses.dataclass
class _Blocking:
    # Indices splitting the matrices, including outer boundaries
    # Example: matmul(A, B) where
    #     A.shape == (m, n), blocking: 2x3
    #     B.shape == (n, p), blocking: 3x1
    # ==> i_partitions == [0, x, m]
    #     j_partitions == [0, p]
    #     k_partitions == [0, y, z, n]
    i_partitions: list[int]
    j_partitions: list[int]
    k_partitions: list[int]


def _find_blocking(
    location_map_a: _BlockLocationMap,
    location_map_b: _BlockLocationMap,
) -> _Blocking:

    i_partitions: list[int] = []
    j_partitions: list[int] = []
    k_partitions: list[int] = []

    def add_to_partitions(indices, partitions):
        start, stop, step = indices

        if step != 1:
            raise RuntimeError('Step other than 1 is not supported')

        partitions.append(start)
        partitions.append(stop)

    for i_indices, k_indices in location_map_a.keys():
        add_to_partitions(i_indices, i_partitions)
        add_to_partitions(k_indices, k_partitions)

    for k_indices, j_indices in location_map_b.keys():
        add_to_partitions(k_indices, k_partitions)
        add_to_partitions(j_indices, j_partitions)

    def to_unique_sorted(partitions):
        if len(partitions) == 0:
            raise RuntimeError('Array has no chunk')

        partitions.sort()

        res = [partitions[0]]
        for x, y in zip(partitions, partitions[1:]):
            if x != y:
                res.append(y)

        return res

    i_partitions = to_unique_sorted(i_partitions)
    j_partitions = to_unique_sorted(j_partitions)
    k_partitions = to_unique_sorted(k_partitions)

    def check_indices(indices, partitions):
        start, stop, _ = indices
        if partitions.index(start) + 1 != partitions.index(stop):
            raise RuntimeError('Inconsistent index mapping')

    for i_indices, k_indices in location_map_a.keys():
        check_indices(i_indices, i_partitions)
        check_indices(k_indices, k_partitions)

    for k_indices, j_indices in location_map_b.keys():
        check_indices(k_indices, k_partitions)
        check_indices(j_indices, j_partitions)

    return _Blocking(i_partitions, j_partitions, k_partitions)


def _make_execution_plan(
    blocking: _Blocking,
    location_map_a: _BlockLocationMap,
    location_map_b: _BlockLocationMap,
) -> _ExecutionPlan:

    i_partitions = blocking.i_partitions
    j_partitions = blocking.j_partitions
    k_partitions = blocking.k_partitions

    plan: _ExecutionPlan = []

    for i_range in zip(i_partitions, i_partitions[1:]):
        for j_range in zip(j_partitions, j_partitions[1:]):
            for k_range in zip(k_partitions, k_partitions[1:]):
                block_a = (i_range + (1,), k_range + (1,))
                block_b = (k_range + (1,), j_range + (1,))

                devices_a = set(location_map_a[block_a].keys())
                devices_b = set(location_map_b[block_b].keys())

                intersection = devices_a & devices_b
                if intersection:
                    # TODO: Pick an execution device that has less work
                    # allocated so far
                    dev = intersection.pop()
                    plan.append((block_a, block_b, dev))
                else:
                    raise RuntimeError(
                        'There is no device that can perform multiplication'
                        f' between block {block_a} and {block_b}')

    return plan


def _convert_to_tuples(
    slices: tuple[slice, ...], shape: tuple[int, ...],
) -> tuple[_SliceIndices, ...]:
    assert len(slices) == len(shape)
    return tuple(s.indices(length) for s, length in zip(slices, shape))


def _convert_to_slices(
    tuples: tuple[_SliceIndices, ...],
) -> tuple[slice, ...]:
    return tuple(slice(*t) for t in tuples)


def _group_by_batch(
    shape: tuple[int, ...], index_map: dict[int, list[tuple[slice, ...]]],
) -> dict[_BatchIdx, _BlockLocationMap]:
    location_maps: dict[_BatchIdx, _BlockLocationMap] = {}

    for dev, idxs in index_map.items():
        for chunk_i, idx in enumerate(idxs):
            idx_tuples = _convert_to_tuples(idx, shape)
            batch_idx, block_idx = idx_tuples[:-2], idx_tuples[-2:]
            block_idx = typing.cast(_BlockIdx, block_idx)

            location_map = location_maps.setdefault(batch_idx, {})
            location = location_map.setdefault(block_idx, {})
            location[dev] = chunk_i

    return location_maps


def _reshape_array_with(
    arr: '_array.DistributedArray',
    f_shape: Callable[[tuple[int,   ...]], tuple[int,   ...]],
    f_idx:   Callable[[tuple[slice, ...]], tuple[slice, ...]],
) -> '_array.DistributedArray':
    def reshape_chunk(chunk: _chunk._Chunk) -> _chunk._Chunk:
        data = chunk.array.reshape(f_shape(chunk.array.shape))
        index = f_idx(chunk.index)
        updates = [(data, f_idx(idx)) for data, idx in chunk.updates]
        return _chunk._Chunk(
            data, chunk.ready, index, updates, chunk.prevent_gc)

    chunks_map = {}
    for dev, chunks in arr._chunks_map.items():
        chunks_map[dev] = [reshape_chunk(chunk) for chunk in chunks]

    shape = f_shape(arr.shape)
    return _array.DistributedArray(
        shape, arr.dtype, chunks_map, arr._mode, arr._comms)


def _prepend_one_to_shape(arr) -> '_array.DistributedArray':
    return _reshape_array_with(
        arr,
        lambda shape: (1,) + shape,
        lambda idx: (slice(None),) + idx)


def _append_one_to_shape(arr) -> '_array.DistributedArray':
    return _reshape_array_with(
        arr,
        lambda shape: shape + (1,),
        lambda idx: idx + (slice(None),))


def _pop_from_shape(arr) -> '_array.DistributedArray':
    assert arr.shape[-1] == 1
    return _reshape_array_with(
        arr,
        lambda shape: shape[:-1],
        lambda idx: idx[:-1])


def _pop_front_from_shape(arr) -> '_array.DistributedArray':
    assert arr.shape[0] == 1
    return _reshape_array_with(
        arr,
        lambda shape: shape[1:],
        lambda idx: idx[1:])


def matmul(
    a: '_array.DistributedArray', b: '_array.DistributedArray',
    out: Optional['_array.DistributedArray'] = None, **kwargs,
) -> '_array.DistributedArray':
    """Matrix multiplication between distributed arrays.

    The arguments must have compatible :attr:`~DistributedArray.shape` and
    :attr:`~DistributedArray.index_map`.

    This operation converts its operands into the replica mode, and compute
    their product in the sum mode.

    Args:
        a, b: Input distributed arrays.
        out (optional): A location into which the result is stored. This option
            is currently not supported.
    Returns:
        The matrix product of the inputs.

    Example:
        >>> A = distributed_array(
        ...     cupy.arange(6).reshape(2, 3),
        ...     make_2d_index_map([0, 2], [0, 1, 3],
        ...                       [[{0}, {1, 2}]]))
        >>> B = distributed_array(
        ...     cupy.arange(12).reshape(3, 4),
        ...     make_2d_index_map([0, 1, 3], [0, 2, 4],
        ...                       [[{0}, {0}],
        ...                        [{1}, {2}]]))
        >>> C = A @ B
        >>> C.mode
        'sum'
        >>> C.all_chunks()
        {0: [array([[0, 0],
                    [0, 3]]),
             array([[0, 0],
                    [6, 9]])],
         1: [array([[20, 23],
                    [56, 65]])],
         2: [array([[26, 29],
                    [74, 83]])]}
        >>> C
        array([[20, 23, 26, 29],
               [56, 68, 80, 92]])

    .. seealso:: :obj:`numpy.matmul`
    """
    if out is not None:
        raise RuntimeError('Argument `out` is not supported')
    for param in ('subok', 'axes', 'axis'):
        if param in kwargs:
            raise RuntimeError(f'Argument `{param}` is not supported')
    if (not isinstance(a, _array.DistributedArray)
            or not isinstance(b, _array.DistributedArray)):
        raise RuntimeError(
            'Mixing a distributed array with a non-distributed array is not'
            ' supported')

    a = a._to_op_mode(_modes.REPLICA)
    b = b._to_op_mode(_modes.REPLICA)

    one_prepended = one_appended = False
    if a.ndim == 1:
        one_prepended = True
        a = _prepend_one_to_shape(a)
    if b.ndim == 1:
        one_appended = True
        b = _append_one_to_shape(b)

    n, m = a.shape[-2:]
    m2, p = b.shape[-2:]
    if m != m2 or a.shape[:-2] != b.shape[:-2]:
        raise ValueError('Shapes are incompatible')

    location_maps_a = _group_by_batch(a.shape, a.index_map)
    location_maps_b = _group_by_batch(b.shape, b.index_map)
    if location_maps_a.keys() != location_maps_b.keys():
        raise RuntimeError('Mismatched batch shapes')

    chunks_map: dict[int, list[_chunk._Chunk]] = {dev: [] for dev in a.devices}
    dtype = None

    for batch_idx in location_maps_a.keys():
        location_map_a = location_maps_a[batch_idx]
        location_map_b = location_maps_b[batch_idx]

        blocking = _find_blocking(location_map_a, location_map_b)
        plan = _make_execution_plan(blocking, location_map_a, location_map_b)

        index_prefix = _convert_to_slices(batch_idx)
        for block_a, block_b, dev in plan:
            loc_a = location_map_a[block_a]
            loc_b = location_map_b[block_b]
            chunk_a = a._chunks_map[dev][loc_a[dev]]
            chunk_b = b._chunks_map[dev][loc_b[dev]]
            chunk_a.flush(_modes.REPLICA)
            chunk_b.flush(_modes.REPLICA)

            index = index_prefix + (slice(*block_a[0]), slice(*block_b[1]))
            with chunk_a.on_ready() as stream:
                stream.wait_event(chunk_b.ready)

                chunk_ab_array = cupy.linalg._product.matmul(
                    chunk_a.array, chunk_b.array, **kwargs)

                chunk_ab = _chunk._Chunk(
                    chunk_ab_array, stream.record(), index,
                    prevent_gc=(chunk_a, chunk_b))
                chunks_map[dev].append(chunk_ab)
                dtype = chunk_ab_array.dtype

    shape = a.shape[:-2] + (n, p)
    res = _array.DistributedArray(
        shape, dtype, chunks_map, _modes.SUM, a._comms)

    if one_prepended:
        res = _pop_front_from_shape(res)
    if one_appended:
        res = _pop_from_shape(res)

    return res


def make_2d_index_map(
    i_partitions: list[int],
    j_partitions: list[int],
    devices: list[list[set[int]]],
) -> dict[int, list[tuple[slice, ...]]]:
    """Create an ``index_map`` for a 2D matrix with a specified blocking.

    Args:
        i_partitions (list of ints): boundaries of blocks on the `i` axis
        j_partitions (list of ints): boundaries of blocks on the `j` axis
        devices (2D list of sets of ints): devices owning each block

    Returns:
        dict from int to array indices: index_map
            Indices for the chunks that devices with designated IDs are going
            to own.

    Example:
        >>> index_map = make_2d_index_map(
        ...     [0, 2, 4], [0, 3, 5],
        ...     [[{0}, {1}],
        ...      [{2}, {0, 1}]])
        >>> pprint(index_map)
        {0: [(slice(0, 2, None), slice(0, 3, None)),
             (slice(2, 4, None), slice(3, 5, None))],
         1: [(slice(0, 2, None), slice(3, 5, None)),
             (slice(2, 4, None), slice(3, 5, None))],
         2: [(slice(2, 4, None), slice(0, 3, None))]}
    """
    assert i_partitions[0] == 0
    assert sorted(set(i_partitions)) == i_partitions
    assert j_partitions[0] == 0
    assert sorted(set(j_partitions)) == j_partitions

    index_map: dict[int, list[tuple[slice, ...]]] = {}
    assert len(devices) == len(i_partitions) - 1
    for i in range(len(devices)):
        assert len(devices[i]) == len(j_partitions) - 1
        for j in range(len(devices[i])):
            i_start = i_partitions[i]
            i_stop = i_partitions[i+1]
            j_start = j_partitions[j]
            j_stop = j_partitions[j+1]

            idx = (slice(i_start, i_stop), slice(j_start, j_stop))

            for dev in devices[i][j]:
                index_map.setdefault(dev, []).append(idx)

    return index_map



def scatter_add(a, slices, value):
    """Adds given values to specified elements of an array.

    It adds ``value`` to the specified elements of ``a``.
    If all of the indices target different locations, the operation of
    :func:`scatter_add` is equivalent to ``a[slices] = a[slices] + value``.
    If there are multiple elements targeting the same location,
    :func:`scatter_add` uses all of these values for addition. On the other
    hand, ``a[slices] = a[slices] + value`` only adds the contribution from one
    of the indices targeting the same location.

    Note that just like an array indexing, negative indices are interpreted as
    counting from the end of an array.

    Also note that :func:`scatter_add` behaves identically
    to :func:`numpy.add.at`.

    Example
    -------
    >>> import cupy
    >>> import cupyx
    >>> a = cupy.zeros((6,), dtype=cupy.float32)
    >>> i = cupy.array([1, 0, 1])
    >>> v = cupy.array([1., 1., 1.])
    >>> cupyx.scatter_add(a, i, v);
    >>> a
    array([1., 2., 0., 0., 0., 0.], dtype=float32)

    Args:
        a (ndarray): An array that gets added.
        slices: It is integer, slices, ellipsis, numpy.newaxis,
            integer array-like, boolean array-like or tuple of them.
            It works for slices used for
            :func:`cupy.ndarray.__getitem__` and
            :func:`cupy.ndarray.__setitem__`.
        v (array-like): Values to increment ``a`` at referenced locations.

    .. note::
        It only supports types that are supported by CUDA's atomicAdd when
        an integer array is included in ``slices``.
        The supported types are ``numpy.float32``, ``numpy.int32``,
        ``numpy.uint32``, ``numpy.uint64`` and ``numpy.ulonglong``.

    .. note::
        :func:`scatter_add` does not raise an error when indices exceed size of
        axes. Instead, it wraps indices.

    .. seealso:: :meth:`numpy.ufunc.at`.

    """
    a.scatter_add(slices, value)


def scatter_max(a, slices, value):
    """Stores a maximum value of elements specified by indices to an array.

    It stores the maximum value of elements in ``value`` array indexed by
    ``slices`` to ``a``. If all of the indices target different locations,
    the operation of :func:`scatter_max` is equivalent to
    ``a[slices] = cupy.maximum(a[slices], value)``.
    If there are multiple elements targeting the same location,
    :func:`scatter_max` stores the maximum of all of these values to the given
    index of ``a``, the initial element of ``a`` is also taken in account.

    Note that just like an array indexing, negative indices are interpreted as
    counting from the end of an array.

    Also note that :func:`scatter_max` behaves identically
    to :func:`numpy.maximum.at`.

    Example
    -------
    >>> import numpy
    >>> import cupy
    >>> a = cupy.zeros((6,), dtype=numpy.float32)
    >>> i = cupy.array([1, 0, 1, 2])
    >>> v = cupy.array([1., 2., 3., -1.])
    >>> cupyx.scatter_max(a, i, v);
    >>> a
    array([2., 3., 0., 0., 0., 0.], dtype=float32)

    Args:
        a (ndarray): An array to store the results.
        slices: It is integer, slices, ellipsis, numpy.newaxis,
            integer array-like, boolean array-like or tuple of them.
            It works for slices used for
            :func:`cupy.ndarray.__getitem__` and
            :func:`cupy.ndarray.__setitem__`.
        v (array-like): An array used for reference.
    """
    a.scatter_max(slices, value)


def scatter_min(a, slices, value):
    """Stores a minimum value of elements specified by indices to an array.

    It stores the minimum value of elements in ``value`` array indexed by
    ``slices`` to ``a``. If all of the indices target different locations,
    the operation of :func:`scatter_min` is equivalent to
    ``a[slices] = cupy.minimum(a[slices], value)``.
    If there are multiple elements targeting the same location,
    :func:`scatter_min` stores the minimum of all of these values to the given
    index of ``a``, the initial element of ``a`` is also taken in account.

    Note that just like an array indexing, negative indices are interpreted as
    counting from the end of an array.

    Also note that :func:`scatter_min` behaves identically
    to :func:`numpy.minimum.at`.

    Example
    -------
    >>> import numpy
    >>> import cupy
    >>> a = cupy.zeros((6,), dtype=numpy.float32)
    >>> i = cupy.array([1, 0, 1, 2])
    >>> v = cupy.array([1., 2., 3., -1.])
    >>> cupyx.scatter_min(a, i, v);
    >>> a
    array([ 0.,  0., -1.,  0.,  0.,  0.], dtype=float32)

    Args:
        a (ndarray): An array to store the results.
        slices: It is integer, slices, ellipsis, numpy.newaxis,
            integer array-like, boolean array-like or tuple of them.
            It works for slices used for
            :func:`cupy.ndarray.__getitem__` and
            :func:`cupy.ndarray.__setitem__`.
        v (array-like): An array used for reference.
    """
    a.scatter_min(slices, value)
